#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#

import logging
import aenum

logger = logging.getLogger(__name__)


# Use abstract base class instead of Mixin to inherit from `aenum.Enum` to make Sphinx
# detect inheritance correctly and create docs for derived enums, including such as
# these in `dev.supercube.constants`. With Mixin approach, module-level enum classes
# are not documented.
class StrEnumBase(aenum.Enum):
    """
    String representation-based equality and lookup.
    """

    def __eq__(self, other):
        return (self is other) or (other.__eq__(str(self)))

    # use only with aenum enums
    @classmethod
    def _missing_value_(cls, value):
        for member in cls:
            if member == value:
                return member

    def __str__(self):
        raise NotImplementedError

    def __hash__(self):
        return hash(str(self))


unique = aenum.unique


class ValueEnum(StrEnumBase):
    """
    Enum with string representation of values used as string representation, and with
    lookup and equality based on this representation.

    Attention: to avoid errors, best use together with `unique` enum decorator.
    """

    def __str__(self):
        return str(self.value)


class NameEnum(StrEnumBase):
    """
    Enum with names used as string representation, and with lookup and equality based on
    this representation.
    """

    # convenience: enables `[]` name-based lookup with enum instances themselves
    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        return self.name


class AutoNumberNameEnum(NameEnum, aenum.AutoNumberEnum):
    """
    Auto-numbered enum with names used as string representation, and with lookup and
    equality based on this representation.
    """

    pass



#  Copyright (c) 2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for .dev sub-package technix
"""
import logging
import threading
from time import sleep

import pytest

from hvl_ccb.dev import (
    TechnixTelnetCommunication,
    Technix,
    TechnixError,
    TechnixSerialCommunication,
)
from hvl_ccb.dev.technix import TechnixSerialCommunicationConfig
from tests.masked_comm import LocalTechnixServer
from tests.masked_comm.serial import TechnixLoopSerialCommunication
from tests.masked_comm.uitls import get_free_tcp_port

logging.basicConfig(level=logging.DEBUG)


@pytest.fixture(scope="function")
def com_telnet():
    host = "127.0.0.1"
    return {
        "host": host,
        "port": get_free_tcp_port(host),
        "timeout": 0.01,
        "wait_sec_read_text_nonempty": 0.01,
        "default_n_attempts_read_text_nonempty": 2,
    }


@pytest.fixture(scope="module")
def com_serial():
    return {
        "port": "loop://?logging=debug",
        "baudrate": 9600,
        "parity": TechnixSerialCommunicationConfig.Parity.NONE,
        "stopbits": TechnixSerialCommunicationConfig.Stopbits.ONE,
        "bytesize": TechnixSerialCommunicationConfig.Bytesize.EIGHTBITS,
        "timeout": 0.01,
        "wait_sec_read_text_nonempty": 0.01,
        "default_n_attempts_read_text_nonempty": 2,
    }


@pytest.fixture(scope="module")
def dev_config_telnet():
    return {
        "max_voltage": 10000,
        "max_current": 1.5,
        "communication_channel": TechnixTelnetCommunication,
        "post_stop_pause_sec": 0.01,
        "register_pulse_time": 0.01,
    }


@pytest.fixture(scope="module")
def dev_config_serial():
    return {
        "max_voltage": 10000,
        "max_current": 1.5,
        "communication_channel": TechnixSerialCommunication,
        "post_stop_pause_sec": 0.01,
        "register_pulse_time": 0.01,
    }


def start_technix_telnet(com, dev_config):
    ts = LocalTechnixServer(port=com["port"], timeout=com["timeout"])
    x = threading.Thread(target=ts.open)
    x.start()
    tex = Technix(com, dev_config)
    x.join()

    return ts, tex


def full_start_devices(com, dev_config):
    ts, tex = start_technix_telnet(com, dev_config)
    tex.start()
    return ts, tex


def test_devices(com_telnet, dev_config_telnet):
    ts, tex = start_technix_telnet(com_telnet, dev_config_telnet)
    assert ts is not None
    assert tex.__class__ is Technix
    tex.stop()
    ts.close()


def test_watchdog(com_telnet, dev_config_telnet):
    ts, tex = full_start_devices(com_telnet, dev_config_telnet)
    sleep(5)

    tex.stop()
    ts.close()


def test_status_byte(com_telnet, dev_config_telnet):
    ts, tex = full_start_devices(com_telnet, dev_config_telnet)

    # Wrong status byte
    with pytest.raises(TechnixError):
        ts.status = 1000
        tex.get_status_byte()

    ts.status = 1
    assert tex.get_status_byte().__repr__() == "StatusByte: 00000001"
    assert tex.get_status_byte().msb_first(0) is None

    tex.stop()
    ts.close()


def test_no_properties(com_telnet, dev_config_telnet):
    ts, tex = start_technix_telnet(com_telnet, dev_config_telnet)

    assert tex.voltage_regulation is None
    assert tex.hv is None
    assert tex.remote is None
    assert tex.inhibit is None

    tex.stop()
    ts.close()


def test_properties(com_telnet, dev_config_telnet):
    ts, tex = full_start_devices(com_telnet, dev_config_telnet)

    ts.status = 0
    tex.get_status_byte()
    assert not tex.voltage_regulation
    assert not tex.hv
    assert tex.remote
    assert tex.inhibit

    ts.status = (1 << 0) + (1 << 3) + (1 << 6) + (1 << 7)
    tex.get_status_byte()
    assert tex.voltage_regulation
    assert tex.hv
    assert not tex.remote
    assert not tex.inhibit

    tex.stop()
    ts.close()


def test_voltage_current(com_telnet, dev_config_telnet):
    ts, tex = full_start_devices(com_telnet, dev_config_telnet)

    assert tex.max_voltage == 10000
    assert tex.max_current == 1.5

    ts.custom_answer = "d1,102"
    tex.voltage = 250
    assert ts.last_request == "d1,102"

    ts.custom_answer = "d2,2730"
    tex.current = 1
    assert ts.last_request == "d2,2730"

    ts.custom_answer = "a12048"
    assert int(tex.voltage) == 5001
    assert ts.last_request == "a1"

    ts.custom_answer = "a23000"
    assert int(tex.current * 1000) == 1098
    assert ts.last_request == "a2"

    with pytest.raises(TechnixError):
        tex.voltage = 1e6
    with pytest.raises(TechnixError):
        tex.current = 1e6

    tex.stop()
    ts.close()


def test_hv_remote(com_telnet, dev_config_telnet):
    ts, tex = full_start_devices(com_telnet, dev_config_telnet)

    with pytest.raises(TechnixError):
        tex.hv = 100
    tex.hv = 1
    tex.hv = True
    tex.hv = 0
    tex.hv = False

    with pytest.raises(TechnixError):
        tex.remote = 100

    with pytest.raises(TechnixError):
        tex.inhibit = 100
    tex.inhibit = True

    tex.stop()
    ts.close()


def test_fault_detection(com_telnet, dev_config_telnet):
    ts, tex = full_start_devices(com_telnet, dev_config_telnet)

    tex.get_status_byte()
    assert tex._real_status
    with pytest.raises(TechnixError):
        ts.status = 127
        tex.get_status_byte()

    ts.close()


def test_no_or_wrong_reply(com_telnet, dev_config_telnet):
    ts, tex = full_start_devices(com_telnet, dev_config_telnet)

    with pytest.raises(TechnixError):
        ts.listen_and_repeat = []
        tex.inhibit = 0

    with pytest.raises(TechnixError):
        ts.custom_answer = "P7,1"
        ts.listen_and_repeat = []
        tex.inhibit = 0

    with pytest.raises(TechnixError):
        ts.custom_answer = "d1,1234"
        assert tex.current == 666

    ts.close()


def start_serial_devices(com_serial, dev_config_serial):
    com = TechnixLoopSerialCommunication(com_serial)
    com.open()

    tex = Technix(com, dev_config_serial)

    com.put_text("P7,0")
    com.put_text("P6,1")
    com.put_text("P6,0")
    com.put_text("E0")  # status byte for the polling thread
    tex.start()
    assert com.get_written() == "P7,0"
    assert com.get_written() == "P6,1"
    assert com.get_written() == "P6,0"
    sleep(0.1)  # time for the polling thread to start
    assert com.get_written() == "E"
    return com, tex


def test_serial(com_serial, dev_config_serial):
    com, tex = start_serial_devices(com_serial, dev_config_serial)

    com.put_text("E0")
    assert tex.get_status_byte().__repr__() == "StatusByte: 00000000"
    assert tex._real_status
    assert com.get_written() == "E"

    com.put_text("P6,1")
    com.put_text("P6,0")
    com.put_text("P7,1")
    tex.stop()
    assert com.get_written() == "P6,1"
    assert com.get_written() == "P6,0"
    assert com.get_written() == "P7,1"

    com.close()


def test_serial_no_answer(com_serial, dev_config_serial):
    com, tex = start_serial_devices(com_serial, dev_config_serial)

    with pytest.raises(TechnixError):
        assert tex.get_status_byte().__repr__() == "StatusByte: 00000000"



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Supercube package with implementation for the old system version from 2015 based on
Siemens WinAC soft-PLC on an industrial 32bit Windows computer.
"""

from .base import (  # noqa: F401
    SupercubeConfiguration,
)
from .typ_a import Supercube2015WithFU  # noqa: F401



#  Copyright (c) 2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Socket to mime a Telnet server for the tests
"""
import logging
import socket
import threading
from abc import abstractmethod
from builtins import super
from time import sleep

logging.basicConfig(level=logging.DEBUG)


class LocalTelnetTestServer:
    """
    Local Telnet Sever for testing
    """

    def __init__(self, port=23, timeout=1):
        self._host = ""  # "127.0.0.1"
        self._port = port
        self._timeout = timeout
        self._s = socket.socket()
        self._client = None
        self._addr = None
        logging.debug("LocalTelnetTestServer created")
        logging.debug("Bind...")
        self._s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._s.bind((self._host, self._port))
        logging.debug("Start listen")
        self._s.listen()

    def __del__(self):
        _s = getattr(self, "_s", None)
        if _s is not None:
            _s.close()
            del _s

    def open(self):
        logging.debug("Accepting?")
        self._s.settimeout(self._timeout)
        self._client, self._addr = self._s.accept()
        logging.debug(f"Accepted: {self._client} with {self._addr}")
        self._client.settimeout(self._timeout)

    def close(self):
        if self._client is not None:
            self._client.close()
            self._client = None

    def put_text(self, text: str, encoding=None):
        if encoding is None:
            self._client.send(text.encode())
        else:
            self._client.send(text.encode(encoding=encoding))

    def get_written(self):
        # logging.debug("Receiving")
        try:
            data = self._client.recv(1024)
            # logging.debug(f"Received: {data}")
            return data.decode().strip()
        except socket.timeout:
            return None


class TechnixMockup:
    def __init__(self):

        self.status: int = 0
        self.keep_running: bool = True
        self.listen_and_repeat = [
            "P5,0",
            "P5,1",
            "P6,0",
            "P6,1",
            "P7,0",
            "P7,1",
            "P8,0",
            "P8,1",
        ]

        self.last_request = ""
        self.custom_answer = ""

        self._x = threading.Thread(target=self.listen_and_answer)

    @abstractmethod
    def get_written(self):
        return ""

    @abstractmethod
    def put_text(self, request):
        pass

    def listen_and_answer(self):
        while self.keep_running:
            request = self.get_written()
            if not request:
                continue

            logging.debug(f"TechnixMockup got: {request}")

            if request == "E":
                self.put_text(f"E{self.status}")
            elif request in self.listen_and_repeat:
                self.put_text(request)
                logging.debug(f"TechnixMockup returned the request: {request}")
            else:
                self.last_request = request
                self.put_text(self.custom_answer)

            sleep(0.01)


class LocalTechnixServer(TechnixMockup):
    def __init__(self, port=4660, timeout=1):
        super().__init__()
        self._ts = LocalTelnetTestServer(port=port, timeout=timeout)

    def open(self):
        self._ts.open()
        self._x.start()

    def close(self):
        self.keep_running = False
        self._x.join()
        self._ts.close()

    def get_written(self):
        return self._ts.get_written()

    def put_text(self, request, encoding=None):
        self._ts.put_text(text=request, encoding=encoding)



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Supercube Typ A module.
"""

import logging
from time import sleep

from opcua.ua import Variant, VariantType

from hvl_ccb.configuration import configdataclass
from . import constants
from .base import (
    Supercube2015Base,
    SupercubeOpcUaCommunication,
    SupercubeOpcUaCommunicationConfig,
)

logger = logging.getLogger(__name__)


@configdataclass
class SupercubeAOpcUaConfiguration(SupercubeOpcUaCommunicationConfig):
    endpoint_name: str = constants.SupercubeOpcEndpoint.A.value  # type: ignore


class SupercubeAOpcUaCommunication(SupercubeOpcUaCommunication):
    @staticmethod
    def config_cls():
        return SupercubeAOpcUaConfiguration


class Supercube2015WithFU(Supercube2015Base):
    """
    Variant A of the Supercube with frequency converter.
    """

    @staticmethod
    def default_com_cls():
        return SupercubeAOpcUaCommunication

    def set_target_voltage(self, volt_v: float) -> None:
        """
        Set the output voltage to a defined value in V.

        :param volt_v: the desired voltage in V
        """

        self.write(
            constants.Power.voltage_target,
            Variant(volt_v / 1000, varianttype=VariantType.Float),
        )

    def get_target_voltage(self) -> float:
        """
        Gets the current setpoint of the output voltage value in V. This is not a
        measured value but is the corresponding function to :meth:`set_target_voltage`.

        :return: the setpoint voltage in V.
        """

        return float(self.read(constants.Power.voltage_target)) * 1000

    def get_max_voltage(self) -> float:
        """
        Reads the maximum voltage of the setup and returns in V.

        :return: the maximum voltage of the setup in V.
        """

        return float(self.read(constants.Power.voltage_max)) * 1000

    def set_slope(self, slope: float) -> None:
        """
        Sets the dV/dt slope of the Supercube frequency converter to a new value in
        V/s.

        :param slope: voltage slope in V/s (0..15'000)
        """

        self.write(
            constants.Power.voltage_slope,
            Variant(slope / 1000, varianttype=VariantType.Float),
        )

    def get_primary_voltage(self) -> float:
        """
        Read the current primary voltage at the output of the frequency converter (
        before transformer).

        :return: primary voltage in V
        """

        return float(self.read(constants.Power.voltage_primary))

    def get_primary_current(self) -> float:
        """
        Read the current primary current at the output of the frequency converter (
        before transformer).

        :return: primary current in A
        """

        return float(self.read(constants.Power.current_primary))

    def get_frequency(self) -> float:
        """
        Read the electrical frequency of the current Supercube setup.

        :return: the frequency in Hz
        """

        return float(self.read(constants.Power.frequency))

    def get_power_setup(self) -> constants.PowerSetup:
        """
        Return the power setup selected in the Supercube's settings.

        :return: the power setup
        """

        return constants.PowerSetup(self.read(constants.Power.setup))

    def get_fso_active(self) -> bool:
        """
        Get the state of the fast switch off functionality. Returns True if it is
        enabled, False otherwise.

        :return: state of the FSO functionality
        """

        return self.read(constants.BreakdownDetection.activated)

    def fso_reset(self) -> None:
        """
        Reset the fast switch off circuitry to go back into normal state and allow to
        re-enable operate mode.
        """

        self.write(constants.BreakdownDetection.reset, True)
        sleep(0.1)
        self.write(constants.BreakdownDetection.reset, False)



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Supercube package with implementation for system versions from 2019 on (new concept
with hard-PLC Siemens S7-1500 as CPU).
"""

from .base import SupercubeConfiguration  # noqa: F401
from .typ_a import SupercubeWithFU  # noqa: F401
from .typ_b import SupercubeB  # noqa: F401



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""

"""

import logging

import libtiepie as ltp
from libtiepie import generator as ltp_gen

from ...utils.enum import NameEnum

from typing import (
    cast,
    Optional,
    Union,
)
from .base import (
    TiePieDeviceType,
    wrap_libtiepie_exception,
    _verify_via_libtiepie,
    _require_dev_handle,
)
from .utils import (
    PublicPropertiesReprMixin,
    _validate_bool,
    _validate_number,
)
from .oscilloscope import TiePieOscilloscope

logger = logging.getLogger(__name__)


class TiePieGeneratorSignalType(NameEnum):
    _init_ = "value description"
    UNKNOWN = ltp.ST_UNKNOWN, "Unknown"
    SINE = ltp.ST_SINE, "Sine"
    TRIANGLE = ltp.ST_TRIANGLE, "Triangle"
    SQUARE = ltp.ST_SQUARE, "Square"
    DC = ltp.ST_DC, "DC"
    NOISE = ltp.ST_NOISE, "Noise"
    ARBITRARY = ltp.ST_ARBITRARY, "Arbitrary"
    PULSE = ltp.ST_PULSE, "Pulse"


class TiePieGeneratorConfigLimits:
    """
    Default limits for generator parameters.
    """

    def __init__(self, dev_gen: ltp_gen.Generator) -> None:
        self.frequency = (0, dev_gen.frequency_max)
        self.amplitude = (0, dev_gen.amplitude_max)
        self.offset = (None, dev_gen.offset_max)


class TiePieGeneratorConfig(PublicPropertiesReprMixin):
    """
    Generator's configuration with cleaning of values in properties setters.
    """

    def __init__(self, dev_gen: ltp_gen.Generator):
        self.dev_gen: ltp_gen.Generator = dev_gen
        self.param_lim: TiePieGeneratorConfigLimits = TiePieGeneratorConfigLimits(
            dev_gen=dev_gen
        )

    def clean_frequency(self, frequency: float) -> float:
        _validate_number("Frequency", frequency, limits=self.param_lim.frequency)
        frequency = _verify_via_libtiepie(self.dev_gen, "frequency", frequency)
        return float(frequency)

    @property
    def frequency(self) -> float:
        return self.dev_gen.frequency

    @frequency.setter
    def frequency(self, frequency: float) -> None:
        frequency = self.clean_frequency(frequency)
        self.dev_gen.frequency = frequency
        logger.info(f"Generator frequency is set to {frequency} Hz.")

    def clean_amplitude(self, amplitude: float) -> float:
        _validate_number(
            "Generator amplitude", amplitude, limits=self.param_lim.amplitude
        )
        amplitude = _verify_via_libtiepie(self.dev_gen, "amplitude", amplitude)
        return float(amplitude)

    @property
    def amplitude(self) -> float:
        return self.dev_gen.amplitude

    @amplitude.setter
    def amplitude(self, amplitude: float) -> None:
        amplitude = self.clean_amplitude(amplitude)
        self.dev_gen.amplitude = amplitude
        logger.info(f"Generator amplitude is set to {amplitude} V.")

    def clean_offset(self, offset: float) -> float:
        _validate_number("Generator offset", offset, limits=self.param_lim.offset)
        offset = _verify_via_libtiepie(self.dev_gen, "offset", offset)
        return float(offset)

    @property
    def offset(self) -> float:
        return self.dev_gen.offset

    @offset.setter
    def offset(self, offset: float) -> None:
        offset = self.clean_offset(offset)
        self.dev_gen.offset = offset
        logger.info(f"Generator offset is set to {offset} V.")

    @staticmethod
    def clean_signal_type(
        signal_type: Union[int, TiePieGeneratorSignalType]
    ) -> TiePieGeneratorSignalType:
        return TiePieGeneratorSignalType(signal_type)

    @property
    def signal_type(self) -> TiePieGeneratorSignalType:
        return TiePieGeneratorSignalType(self.dev_gen.signal_type)

    @signal_type.setter
    def signal_type(self, signal_type: Union[int, TiePieGeneratorSignalType]) -> None:
        self.dev_gen.signal_type = self.clean_signal_type(signal_type).value
        logger.info(f"Signal type is set to {signal_type}.")

    @staticmethod
    def clean_enabled(enabled: bool) -> bool:
        _validate_bool("channel enabled", enabled)
        return enabled

    @property
    def enabled(self) -> bool:
        return self.dev_gen.enabled

    @enabled.setter
    def enabled(self, enabled: bool) -> None:
        self.dev_gen.enabled = self.clean_enabled(enabled)
        if enabled:
            msg = "enabled"
        else:
            msg = "disabled"
        logger.info(f"Generator is set to {msg}.")


class TiePieGeneratorMixin:
    """
    TiePie Generator sub-device.

    A wrapper for the `libtiepie.generator.Generator` class. To be mixed in with
    `TiePieOscilloscope` base class.
    """

    def __init__(self, com, dev_config):
        super().__init__(com, dev_config)

        self._gen: Optional[ltp_gen.Generator] = None

        self.config_gen: Optional[TiePieGeneratorConfig] = None
        """
        Generator's dynamical configuration.
        """

    @_require_dev_handle(TiePieDeviceType.GENERATOR)
    def _gen_config_setup(self) -> None:
        """
        Setup dynamical configuration for the connected generator.
        """
        self.config_gen = TiePieGeneratorConfig(
            dev_gen=self._gen,
        )

    def _gen_config_teardown(self) -> None:
        self.config_gen = None

    def _gen_close(self) -> None:
        if self._gen is not None:
            del self._gen
            self._gen = None

    def start(self) -> None:
        """
        Start the Generator.
        """
        super().start()  # type: ignore
        logger.info("Starting generator")

        self._gen = cast(TiePieOscilloscope, self)._get_device_by_serial_number(
            TiePieDeviceType.GENERATOR
        )
        self._gen_config_setup()

    @wrap_libtiepie_exception
    def stop(self) -> None:
        """
        Stop the generator.
        """
        logger.info("Stopping generator")

        self._gen_config_teardown()
        self._gen_close()

        super().stop()  # type: ignore

    @wrap_libtiepie_exception
    @_require_dev_handle(TiePieDeviceType.GENERATOR)
    def generator_start(self):
        """
        Start signal generation.
        """
        self._gen.start()
        logger.info("Starting signal generation")

    @wrap_libtiepie_exception
    @_require_dev_handle(TiePieDeviceType.GENERATOR)
    def generator_stop(self):
        """
        Stop signal generation.
        """
        self._gen.stop()
        logger.info("Stopping signal generation")



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""

"""
import logging

from ...utils.typing import Number

from collections import Sequence

from typing import (
    cast,
    Optional,
    Tuple,
    Type,
    Union,
)

logger = logging.getLogger(__name__)


def _validate_number(
    x_name: str,
    x: object,
    limits: Tuple = (None, None),
    number_type: Union[Type[Number], Tuple[Type[Number], ...]] = (int, float),
) -> None:
    """
    Validate if given input `x` is a number of given `number_type` type, with value
    between given `limits[0]` and `limits[1]` (inclusive), if not `None`.

    :param x_name: string name of the validate input, use for the error message
    :param x: an input object to validate as number of given type within given range
    :param limits: [lower, upper] limit, with `None` denoting no limit: [-inf, +inf]
    :param number_type: expected type or tuple of types of a number,
    by default `(int, float)`
    :raises TypeError: when the validated input does not have expected type
    :raises ValueError: when the validated input has correct number type but is not
        within given range
    """
    if limits is None:
        limits = (None, None)
    msg = None
    err_cls: Optional[Type[Exception]] = None
    if not isinstance(number_type, Sequence):
        number_type = (number_type,)
    if not isinstance(x, number_type):
        msg = (
            f"{x_name} = {x} has to be of type "
            f"{' or '.join(nt.__name__ for nt in number_type)}"
        )
        err_cls = TypeError
    elif not (
        (limits[0] is None or cast(Number, x) >= limits[0])
        and (limits[1] is None or cast(Number, x) <= limits[1])
    ):
        if limits[0] is None:
            suffix = f"less or equal than {limits[1]}"
        elif limits[1] is None:
            suffix = f"greater or equal than {limits[0]}"
        else:
            suffix = f"between {limits[0]} and {limits[1]} inclusive"
        msg = f"{x_name} = {x} has to be " + suffix
        err_cls = ValueError
    if err_cls is not None:
        logger.error(msg)
        raise err_cls(msg)


def _validate_bool(x_name: str, x: object) -> None:
    """
    Validate if given input `x` is a `bool`.

    :param x_name: string name of the validate input, use for the error message
    :param x: an input object to validate as boolean
    :raises TypeError: when the validated input does not have boolean type
    """
    if not isinstance(x, bool):
        msg = f"{x_name} = {x} has to of type bool"
        logger.error(msg)
        raise TypeError(msg)


class PublicPropertiesReprMixin:
    """General purpose utility mixin that overwrites object representation to a one
    analogous to `dataclass` instances, but using public properties and their values
    instead of `fields`.
    """

    def _public_properties_gen(self):
        """
        Generator that returns instance's properties names and their values,
        for properties that do not start with `"_"`

        :return: attribute name and value tuples
        """
        for name in dir(self):
            if (
                not name.startswith("_")
                and hasattr(self.__class__, name)
                and isinstance(getattr(self.__class__, name), property)
            ):
                yield name, getattr(self, name)

    def __repr__(self):
        attrs = ", ".join(
            [f"{name}={value!r}" for name, value in self._public_properties_gen()]
        )
        return f"{self.__class__.__qualname__}({attrs})"



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for Enum utils.
"""

import pytest

from hvl_ccb.utils.enum import (
    AutoNumberNameEnum,
    NameEnum,
    StrEnumBase,
    ValueEnum,
    unique,
)


def test_strenumbase():
    @unique
    class E(StrEnumBase):
        A = "a"

    with pytest.raises(NotImplementedError):
        E.A == "a"


def test_valueenum():

    with pytest.raises(ValueError):
        @unique
        class F(ValueEnum):
            A = "a"
            B = "a"

    @unique
    class E(ValueEnum):
        A = "a"
        B = "b"

    a = E("a")
    assert hash(a) == hash(a.value)
    assert a.value == a
    assert a == "a"
    assert str(a) == "a"
    assert a == E.A

    assert E["A"] is a
    with pytest.raises(KeyError):
        E["a"]
    with pytest.raises(KeyError):
        E[a]

    assert a != 0
    assert a != 1

    b = E("b")
    assert hash(b) == hash(b.value)
    assert a != b
    assert a != "b"
    assert a != E.B


def test_nameenum():

    class E(NameEnum):
        _init_ = 'custom_name'
        a = 2
        b = 4

    a = E("a")
    assert hash(a) == hash(a.name)
    assert a.name == a
    assert a == "a"
    assert str(a) == "a"
    assert a == E.a

    assert a != 2
    assert a.custom_name == 2

    assert E["a"] is a
    assert E[a] is a

    b = E("b")
    assert hash(b) == hash(b.name)
    assert a != b
    assert a != "b"
    assert a != E.b


def test_autonumbernameenum():

    class E(AutoNumberNameEnum):
        a = ()
        b = ()

    a = E("a")
    assert hash(a) == hash(a.name)
    assert a.name == a
    assert a == "a"
    assert str(a) == "a"
    assert a == E.a

    assert a != 0
    assert a != 1

    assert E["a"] is a
    assert E[a] is a

    b = E("b")
    assert hash(b) == hash(b.name)
    assert a != b
    assert a != "b"
    assert a != E.b



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Test module for the Supercube base device class in the old 2015 version.
"""

import pytest

from hvl_ccb.dev.supercube2015 import constants
from hvl_ccb.dev.supercube2015.base import Supercube2015Base
from tests.opctools import DemoServer


@pytest.fixture(scope='module')
def dev_config():
    return {
        'namespace_index': 7,
    }


@pytest.fixture(scope='module')
def com_config():
    return {
        'host': 'localhost',
        'port': 14567,
        'endpoint_name': '',
        'update_period': 10,
    }


@pytest.fixture(scope='module')
def demo_opcua_server(dev_config):
    opcua_server = DemoServer(dev_config['namespace_index'], 'B', 14567)
    opcua_server.start()

    # add socket nodes
    for socket in constants.GeneralSockets:
        opcua_server.add_var(socket, False, True)

    # add support input and output nodes
    for support in constants.GeneralSupport:
        opcua_server.add_var(support, False, True)

    for safety in constants.Safety:
        opcua_server.add_var(safety, False, True)

    for stick in constants.EarthingStick:
        opcua_server.add_var(stick, False, True)

    opcua_server.add_var(constants.Errors.stop_number, 0, False)
    opcua_server.add_var(constants.Errors.quit, False, True)

    opcua_server.add_var(constants.BreakdownDetection.activated, False, False)
    opcua_server.add_var(constants.BreakdownDetection.triggered, False, False)

    yield opcua_server

    opcua_server.stop()


@pytest.fixture(scope='module')
def cube(com_config, dev_config, demo_opcua_server: DemoServer):
    cube = Supercube2015Base(com_config, dev_config)
    cube.start()
    yield cube
    cube.stop()


def test_constants():
    assert constants.AlarmText.Alarm0 == 'No Alarm.'
    assert constants.AlarmText.get(-1) == constants.AlarmText.not_defined


def test_error_datachange(cube: Supercube2015Base, demo_opcua_server: DemoServer):
    demo_opcua_server.set_var(constants.Errors.stop_number, 3)


def test_set_remote_control(cube: Supercube2015Base, demo_opcua_server: DemoServer):
    with pytest.raises(NotImplementedError):
        cube.set_remote_control(False)


def test_get_support_input(cube: Supercube2015Base):
    assert cube.get_support_input(1, 1) is False


def test_support_output(cube: Supercube2015Base):
    cube.set_support_output(1, 1, False)
    assert cube.get_support_output(1, 1) is False
    cube.set_support_output(1, 1, True)
    assert cube.get_support_output(1, 1) is True


def test_set_support_output_impulse(cube: Supercube2015Base):
    cube.set_support_output_impulse(2, 2, 0.01, True)


def test_get_t13_socket(cube: Supercube2015Base):
    assert cube.get_t13_socket(1) is False
    cube.set_t13_socket(1, True)
    assert cube.get_t13_socket(1) is True

    with pytest.raises(ValueError):
        cube.get_t13_socket(4)


def test_set_t13_socket(cube: Supercube2015Base):
    cube.set_t13_socket(1, True)

    with pytest.raises(ValueError):
        cube.set_t13_socket(4, False)

    with pytest.raises(ValueError):
        cube.set_t13_socket(1, 'on')


def test_get_cee16(cube: Supercube2015Base):
    cube.set_cee16_socket(False)
    assert cube.get_cee16_socket() is False
    cube.set_cee16_socket(True)
    assert cube.get_cee16_socket() is True


def test_set_cee16(cube: Supercube2015Base):
    cube.set_cee16_socket(True)
    cube.set_cee16_socket(False)

    with pytest.raises(ValueError):
        cube.set_cee16_socket(1)

    with pytest.raises(ValueError):
        cube.set_cee16_socket('on')


def test_get_status(cube: Supercube2015Base, demo_opcua_server: DemoServer):
    demo_opcua_server.set_var(constants.Safety.status_green, True)
    demo_opcua_server.set_var(constants.Safety.status_ready_for_red, False)
    assert cube.get_status() == constants.SafetyStatus.GreenNotReady

    demo_opcua_server.set_var(constants.Safety.status_ready_for_red, True)
    assert cube.get_status() == constants.SafetyStatus.GreenReady

    demo_opcua_server.set_var(constants.Safety.status_red, True)
    demo_opcua_server.set_var(constants.Safety.status_green, False)
    assert cube.get_status() == constants.SafetyStatus.RedReady

    demo_opcua_server.set_var(constants.Safety.switchto_operate, True)
    assert cube.get_status() == constants.SafetyStatus.RedOperate

    demo_opcua_server.set_var(constants.Safety.status_error, True)
    demo_opcua_server.set_var(constants.Safety.switchto_operate, False)
    demo_opcua_server.set_var(constants.Safety.status_red, False)
    assert cube.get_status() == constants.SafetyStatus.Error

    demo_opcua_server.set_var(constants.Safety.status_error, False)
    demo_opcua_server.set_var(constants.BreakdownDetection.triggered, True)
    demo_opcua_server.set_var(constants.BreakdownDetection.activated, True)
    assert cube.get_status() == constants.SafetyStatus.QuickStop
    demo_opcua_server.set_var(constants.BreakdownDetection.triggered, False)
    demo_opcua_server.set_var(constants.BreakdownDetection.activated, False)


def test_ready(cube: Supercube2015Base, demo_opcua_server: DemoServer):
    demo_opcua_server.set_var(constants.Safety.status_ready_for_red, True)
    demo_opcua_server.set_var(constants.Safety.status_green, True)
    demo_opcua_server.set_var(constants.Safety.status_red, False)
    demo_opcua_server.set_var(constants.Safety.status_error, False)
    demo_opcua_server.set_var(constants.Safety.switchto_operate, False)
    cube.ready(True)
    cube.ready(False)


def test_operate(cube: Supercube2015Base, demo_opcua_server: DemoServer):
    demo_opcua_server.set_var(constants.Safety.status_green, False)
    demo_opcua_server.set_var(constants.Safety.status_red, True)
    demo_opcua_server.set_var(constants.Safety.switchto_operate, False)
    demo_opcua_server.set_var(constants.Safety.status_ready_for_red, True)
    demo_opcua_server.set_var(constants.Safety.status_error, False)
    cube.operate(True)
    cube.operate(False)


def test_get_measurement_ratio(cube: Supercube2015Base, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(constants.MeasurementsDividerRatio.input_1, 123.4, False)
    assert cube.get_measurement_ratio(1) == 123.4


def test_get_measurement_voltage(cube: Supercube2015Base,
                                 demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(constants.MeasurementsScaledInput.input_1, 110_000, False)
    assert cube.get_measurement_voltage(1) == 110_000.0


def test_get_earthing_status(cube: Supercube2015Base, demo_opcua_server: DemoServer):
    demo_opcua_server.set_var(constants.EarthingStick.status_1_connected, False)
    assert cube.get_earthing_status(1) == constants.EarthingStickStatus.inactive

    demo_opcua_server.set_var(constants.EarthingStick.status_1_connected, True)
    demo_opcua_server.set_var(constants.EarthingStick.status_1_closed, True)
    demo_opcua_server.set_var(constants.EarthingStick.status_1_open, False)
    assert cube.get_earthing_status(1) == constants.EarthingStickStatus.closed

    demo_opcua_server.set_var(constants.EarthingStick.status_1_closed, False)
    assert cube.get_earthing_status(1) == constants.EarthingStickStatus.error

    demo_opcua_server.set_var(constants.EarthingStick.status_1_open, True)
    assert cube.get_earthing_status(1) == constants.EarthingStickStatus.open


def test_earthing_manual(cube: Supercube2015Base, demo_opcua_server: DemoServer):
    demo_opcua_server.set_var(constants.Safety.status_red, False)

    assert cube.get_earthing_manual(1) is False
    cube.set_earthing_manual(1, True)
    assert cube.get_earthing_manual(1) is True

    demo_opcua_server.set_var(constants.Safety.switchto_operate, True)
    demo_opcua_server.set_var(constants.Safety.status_red, True)
    demo_opcua_server.set_var(constants.Safety.status_green, False)
    demo_opcua_server.set_var(constants.Safety.status_error, False)
    demo_opcua_server.set_var(constants.Safety.status_ready_for_red, True)
    with pytest.raises(RuntimeError):
        cube.set_earthing_manual(1, True)


def test_quit_error(cube: Supercube2015Base, demo_opcua_server: DemoServer):
    cube.quit_error()
    assert demo_opcua_server.get_var(constants.Errors.quit) is False


def test_get_door_status(cube: Supercube2015Base, demo_opcua_server: DemoServer):
    with pytest.raises(NotImplementedError):
        cube.get_door_status(1)



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""Communication protocols subpackage."""

from .base import (  # noqa: F401
    CommunicationProtocol,
    NullCommunicationProtocol,
)

try:
    from .labjack_ljm import (  # noqa: F401
        LJMCommunication,
        LJMCommunicationConfig,
        LJMCommunicationError,
    )
except (ImportError, ModuleNotFoundError):
    import warnings

    warnings.warn("To use libtiepie library install hvl with command "
                  "`pip install hvl_ccb[tiepie]`.")

from .modbus_tcp import (  # noqa: F401
    ModbusTcpCommunication,
    ModbusTcpConnectionFailedException,
    ModbusTcpCommunicationConfig,
)
from .opc import (  # noqa: F401
    OpcUaCommunication,
    OpcUaCommunicationConfig,
    OpcUaCommunicationIOError,
    OpcUaCommunicationTimeoutError,
    OpcUaSubHandler,
)
from .telnet import (  # noqa: F401
    TelnetCommunication,
    TelnetCommunicationConfig,
    TelnetError,
)
from .serial import (  # noqa: F401
    SerialCommunication,
    SerialCommunicationConfig,
    SerialCommunicationIOError,
)
from .visa import (  # noqa: F401
    VisaCommunication,
    VisaCommunicationError,
    VisaCommunicationConfig,
)



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
This module is a wrapper around libtiepie Oscilloscope devices; see
https://www.tiepie.com/en/libtiepie-sdk .

The device classes adds simplifications for starting of the device (using serial
number) and managing mutable configuration of both the device and oscilloscope's
channels. This includes extra validation and typing hints support.

To install libtiepie on Windows:
The installation of the Python bindings "python-libtiepie" is done automatically
with the dependencies of the hvl_ccb. The additional DLL for Windows is included in
that package.

On a Linux-system additional libraries have to be installed; see
https://www.tiepie.com/en/libtiepie-sdk/linux .

On a Windows system, if you encounter an :code:`OSError` like this::

    ...
        self._handle = _dlopen(self._name, mode)
    OSError: [WinError 126] The specified module could not be found

most likely the python-libtiepie package was installed in your :code:`site-packages/`
directory as a :code:`python-libtiepie-*.egg` file via :code:`python setup.py
install` or :code:`python setup.py develop` command. In this case uninstall the
library and re-install it using :code:`pip`::

    $ pip uninstall python-libtiepie
    $ pip install python-libtiepie

This should create :code:`libtiepie/` folder. Alternatively, manually move the folder
:code:`libtiepie/` from inside of the :code:`.egg` archive file to the containing it
:code:`site-packages/` directory (PyCharm's Project tool window supports reading and
extracting from :code:`.egg` archives).

"""



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Supercube Typ B module.
"""

import logging
from hvl_ccb.configuration import configdataclass
from . import constants
from .base import (
    SupercubeBase,
    SupercubeOpcUaCommunication,
    SupercubeOpcUaCommunicationConfig,
)

logger = logging.getLogger(__name__)


@configdataclass
class SupercubeBOpcUaConfiguration(SupercubeOpcUaCommunicationConfig):
    endpoint_name: str = constants.SupercubeOpcEndpoint.B.value  # type: ignore


class SupercubeBOpcUaCommunication(SupercubeOpcUaCommunication):
    @staticmethod
    def config_cls():
        return SupercubeBOpcUaConfiguration


class SupercubeB(SupercubeBase):
    """
    Variant B of the Supercube without frequency converter but external safety switches.
    """

    @staticmethod
    def default_com_cls():
        return SupercubeBOpcUaCommunication



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Communication protocol for modbus TCP ports. Makes use of the
`pymodbus <https://pymodbus.readthedocs.io/en/latest/>`_ library.
"""

import logging
from typing import List, Union

from IPy import IP
from pymodbus.client.sync import ModbusTcpClient
from pymodbus.constants import Defaults as ModbusDefaults
from pymodbus.exceptions import ConnectionException

from .base import CommunicationProtocol
from ..configuration import configdataclass

logger = logging.getLogger(__name__)


class ModbusTcpConnectionFailedException(ConnectionException):
    """
    Exception raised when the connection failed.
    """

    pass


@configdataclass
class ModbusTcpCommunicationConfig:
    """
    Configuration dataclass for :class:`ModbusTcpCommunication`.
    """

    #: Host is the IP address of the connected device.
    host: str

    #: Unit number to be used when connecting with Modbus/TCP. Typically this is used
    #: when connecting to a relay having Modbus/RTU-connected devices.
    unit: int

    #: TCP port
    port: int = ModbusDefaults.Port

    def clean_values(self):
        # host, raises ValueError on its own if not suitable
        IP(self.host)


class ModbusTcpCommunication(CommunicationProtocol):
    """
    Implements the Communication Protocol for modbus TCP.
    """

    def __init__(self, configuration):
        """Constructor for modbus"""
        super().__init__(configuration)

        # create the modbus port specified in the configuration
        logger.debug(
            f"Create ModbusTcpClient with host: {self.config.host}, "
            f"Port: {self.config.port}, Unit: {self.config.unit}"
        )
        self.client = ModbusTcpClient(
            self.config.host, port=self.config.port, unit=self.config.unit
        )

    @staticmethod
    def config_cls():
        return ModbusTcpCommunicationConfig

    def open(self) -> None:
        """
        Open the Modbus TCP connection.

        :raises ModbusTcpConnectionFailedException: if the connection fails.
        """

        # open the port
        logger.debug("Open Modbus TCP Port.")

        with self.access_lock:
            if not self.client.connect():
                raise ModbusTcpConnectionFailedException

    def close(self):
        """
        Close the Modbus TCP connection.
        """

        # close the port
        logger.debug("Close Modbus TCP Port.")

        with self.access_lock:
            self.client.close()

    def write_registers(self, address: int, values: Union[List[int], int]):
        """
        Write values from the specified address forward.

        :param address: address of the first register
        :param values: list with all values
        """

        logger.debug(f"Write registers {address} with values {values}")

        with self.access_lock:
            try:
                self.client.write_registers(
                    address=address, values=values, unit=self.config.unit
                )
            except ConnectionException as e:
                logger.error(str(e), exc_info=e)
                raise ModbusTcpConnectionFailedException from e

    def read_holding_registers(self, address: int, count: int) -> List[int]:
        """
        Read specified number of register starting with given address and return
        the values from each register.

        :param address: address of the first register
        :param count: count of registers to read
        :return: list of `int` values
        """

        logger.debug(f"Read holding registers {address} with count {count}.")

        with self.access_lock:
            try:
                registers = self.client.read_holding_registers(
                    address=address, count=count
                ).registers
            except ConnectionException as e:
                logger.error(str(e), exc_info=e)
                raise ModbusTcpConnectionFailedException from e

        logger.debug(f"Returned holding registers {address}: {registers}")

        return registers

    def read_input_registers(self, address: int, count: int) -> List[int]:
        """
        Read specified number of register starting with given address and return
        the values from each register in a list.

        :param address: address of the first register
        :param count: count of registers to read
        :return: list of `int` values
        """

        logger.debug(f"Read input registers {address} with count {count}.")

        with self.access_lock:
            try:
                registers = self.client.read_input_registers(
                    address=address, count=count
                ).registers
            except ConnectionException as e:
                logger.error(str(e), exc_info=e)
                raise ModbusTcpConnectionFailedException from e

        logger.debug(f"Returned input registers {address}: {registers}")

        return registers



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
from .labjack_ljm import MaskedLJMCommunication  # noqa: F401
from .telnet_mockup import LocalTelnetTestServer, LocalTechnixServer  # noqa: F401
from .visa import MaskedVisaCommunication  # noqa: F401



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""Devices subpackage."""

import sys

from .base import (  # noqa: F401
    Device,
    DeviceExistingException,
    DeviceSequenceMixin,
    DeviceFailuresException,
    SingleCommDevice,
)
from .crylas import (  # noqa: F401
    CryLasLaser,
    CryLasLaserConfig,
    CryLasLaserSerialCommunication,
    CryLasLaserSerialCommunicationConfig,
    CryLasLaserError,
    CryLasLaserNotReadyError,
    CryLasAttenuator,
    CryLasAttenuatorConfig,
    CryLasAttenuatorSerialCommunication,
    CryLasAttenuatorSerialCommunicationConfig,
    CryLasAttenuatorError,
)
from .ea_psi9000 import (  # noqa: F401
    PSI9000,
    PSI9000Config,
    PSI9000VisaCommunication,
    PSI9000VisaCommunicationConfig,
    PSI9000Error,
)
from .fug import (  # noqa: F401
    FuG,
    FuGConfig,
    FuGSerialCommunication,
    FuGSerialCommunicationConfig,
    FuGError,
    FuGErrorcodes,
    FuGDigitalVal,
    FuGTerminators,
    FuGPolarities,
    FuGReadbackChannels,
    FuGMonitorModes,
    FuGRampModes,
)
from .heinzinger import (  # noqa: F401
    HeinzingerDI,
    HeinzingerPNC,
    HeinzingerConfig,
    HeinzingerPNCError,
    HeinzingerPNCMaxVoltageExceededException,
    HeinzingerPNCMaxCurrentExceededException,
    HeinzingerPNCDeviceNotRecognizedException,
    HeinzingerSerialCommunication,
    HeinzingerSerialCommunicationConfig,
)

try:
    from .labjack import (  # noqa: F401
        LabJack,
        LabJackError,
        LabJackIdentifierDIOError,
    )
except (ImportError, ModuleNotFoundError):
    import warnings

    warnings.warn("To use labjack library install hvl with command "
                  "`pip install hvl_ccb[labjack]`.")

from .mbw973 import (  # noqa: F401
    MBW973,
    MBW973Config,
    MBW973ControlRunningException,
    MBW973PumpRunningException,
    MBW973Error,
    MBW973SerialCommunication,
    MBW973SerialCommunicationConfig,
)
from .newport import (  # noqa: F401
    NewportSMC100PP,
    NewportSMC100PPConfig,
    NewportStates,
    NewportSMC100PPSerialCommunication,
    NewportSMC100PPSerialCommunicationConfig,
    NewportConfigCommands,
    NewportMotorError,
    NewportControllerError,
    NewportSerialCommunicationError,
    NewportUncertainPositionError,
    NewportMotorPowerSupplyWasCutError,
)
from .pfeiffer_tpg import (  # noqa: F401
    PfeifferTPG,
    PfeifferTPGConfig,
    PfeifferTPGSerialCommunication,
    PfeifferTPGSerialCommunicationConfig,
    PfeifferTPGError,
)
from .rs_rto1024 import (  # noqa: F401
    RTO1024,
    RTO1024Error,
    RTO1024Config,
    RTO1024VisaCommunication,
    RTO1024VisaCommunicationConfig,
)
from .se_ils2t import (  # noqa: F401
    ILS2T,
    ILS2TConfig,
    ILS2TException,
    ILS2TModbusTcpCommunication,
    ILS2TModbusTcpCommunicationConfig,
    IoScanningModeValueError,
    ScalingFactorValueError,
)
from .sst_luminox import (  # noqa: F401
    Luminox,
    LuminoxConfig,
    LuminoxSerialCommunication,
    LuminoxSerialCommunicationConfig,
    LuminoxMeasurementType,
    LuminoxMeasurementTypeError,
    LuminoxOutputMode,
    LuminoxOutputModeError,
)
from .technix import (  # noqa: F401
    Technix,
    TechnixError,
    TechnixTelnetCommunication,
    TechnixSerialCommunication,
)
from .visa import (  # noqa: F401
    VisaDevice,
    VisaDeviceConfig,
)

if sys.platform == 'darwin':
    import warnings

    warnings.warn("libtiepie is not available for Darwin OSs")
else:
    try:
        from .tiepie.generator import (  # noqa: F401
            TiePieGeneratorConfig,
            TiePieGeneratorSignalType,
            TiePieGeneratorMixin,
        )
        from .tiepie.i2c import (  # noqa: F401
            TiePieI2CHostConfig,
            TiePieI2CHostMixin,
        )
        from .tiepie.oscilloscope import (  # noqa: F401
            TiePieOscilloscope,
            TiePieOscilloscopeConfig,
            TiePieOscilloscopeResolution,
            TiePieOscilloscopeAutoResolutionModes,
        )
        from .tiepie.base import (  # noqa: F401
            TiePieDeviceConfig,
            TiePieDeviceType,
            TiePieError,
            get_device_by_serial_number,
            _verify_via_libtiepie,
        )
        from .tiepie.utils import _validate_number  # noqa: F401
        from .tiepie.channel import (  # noqa: F401
            TiePieOscilloscopeChannelCoupling,
            TiePieOscilloscopeChannelConfig,
            TiePieOscilloscopeRange,
            TiePieOscilloscopeTriggerKind,
            TiePieOscilloscopeTriggerLevelMode,
        )
        from .tiepie.device import (  # noqa: F401
            TiePieHS5,
            TiePieHS6,
            TiePieWS5,
        )
    except (ImportError, ModuleNotFoundError):
        import warnings

        warnings.warn("To use libtiepie library install hvl with command "
                      "`pip install hvl_ccb[tiepie]`.")



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Test module for the Supercube base device class.
"""

import logging
from time import sleep
from typing import List

import pytest
from pytest_mock import MockerFixture

from hvl_ccb.comm import OpcUaCommunicationIOError, OpcUaCommunicationTimeoutError
from hvl_ccb.dev.supercube import constants
from hvl_ccb.dev.supercube.base import (
    SupercubeBase,
    SupercubeConfiguration,
    SupercubeEarthingStickOperationError,
    SupercubeOpcUaCommunication,
)
from tests.opctools import DemoServer


@pytest.fixture(scope="module")
def dev_config():
    return {
        "namespace_index": 3,
        "polling_delay_sec": 0.05,
        "polling_interval_sec": 0.01,
    }


@pytest.fixture(scope="module")
def com_config():
    return {
        "host": "localhost",
        "port": 14123,
        "endpoint_name": "",
        "update_period": 10,
        "wait_timeout_retry_sec": 0.01,
        "max_timeout_retry_nr": 3,
    }


@pytest.fixture(scope="function")
def demo_opcua_server(dev_config):
    opcua_server = DemoServer(dev_config["namespace_index"], "B", 14123)
    opcua_server.start()

    # add socket nodes
    for socket in constants.GeneralSockets:
        opcua_server.add_var(socket, False, True)

    # add support input and output nodes
    for support in constants.GeneralSupport:
        opcua_server.add_var(support, False, True)

    for i, alarm in enumerate(constants.Alarms):
        opcua_server.add_var(alarm, bool(i % 2), False)

    # add remote control vars
    opcua_server.add_var(constants.OpcControl.active, False, True)
    opcua_server.add_var(constants.OpcControl.live, False, True)

    yield opcua_server

    opcua_server.stop()


@pytest.fixture(scope="function")
def opened_supercube_com(
    com_config, dev_config, demo_opcua_server: DemoServer,
):
    opc_comm = SupercubeOpcUaCommunication(com_config)
    opc_comm.open()
    yield opc_comm
    opc_comm.close()


def _get_logged_messages(
    caplog, expected_logger_name: str, expected_level: int,
) -> List[str]:
    return [
        message
        for logger_name, level, message in caplog.record_tuples
        if logger_name == expected_logger_name and level == expected_level
    ]


def test_supercube_subscription_handler_datachange_safetystatus(
    opened_supercube_com: SupercubeOpcUaCommunication,
    dev_config: dict,
    demo_opcua_server: DemoServer,
    caplog,  # pytest._logging.LogCaptureFixture
):
    ns_index = dev_config["namespace_index"]

    caplog.set_level(logging.INFO)

    # test a status datachange
    demo_opcua_server.add_var(
        constants.Safety.status, int(constants.SafetyStatus.Initializing), True,
    )
    opened_supercube_com.init_monitored_nodes(str(constants.Safety.status), ns_index)
    sleep(0.05)

    demo_opcua_server.set_var(
        constants.Safety.status, int(constants.SafetyStatus.GreenNotReady),
    )
    sleep(0.05)

    # FIXME: does not work in gitlab CI
    # check presence of log messages in a correct order
    # assert [
    #     f"Safety: {str(constants.SafetyStatus.Initializing)}",
    #     f"Safety: {str(constants.SafetyStatus.GreenNotReady)}",
    # ] == _get_logged_messages(caplog, "hvl_ccb.dev.supercube.base", logging.INFO)


def test_supercube_subscription_handler_datachange_earthingstickstatus(
    opened_supercube_com: SupercubeOpcUaCommunication,
    dev_config: dict,
    demo_opcua_server: DemoServer,
    caplog,  # pytest._logging.LogCaptureFixture
):
    ns_index = dev_config["namespace_index"]

    caplog.set_level(logging.INFO)

    # test an earthing stick status datachange
    ess_tuple = constants.EarthingStick.statuses()
    for ess in ess_tuple:
        demo_opcua_server.add_var(
            ess, int(constants.EarthingStickStatus.inactive), True,
        )
    opened_supercube_com.init_monitored_nodes(ess_tuple, ns_index)
    sleep(0.05)

    for ess in ess_tuple:
        demo_opcua_server.set_var(ess, int(constants.EarthingStickStatus.closed))
    sleep(0.05)

    # FIXME: does not work in gitlab CI
    # check presence of log messages in a correct order
    # assert (
    #     [
    #         f"Earthing {ess.number}: {str(constants.EarthingStickStatus.inactive)}"
    #         for ess in ess_tuple
    #     ]
    #     + [
    #         f"Earthing {ess.number}: {str(constants.EarthingStickStatus.closed)}"
    #         for ess in ess_tuple
    #     ]
    # ) == _get_logged_messages(caplog, "hvl_ccb.dev.supercube.base", logging.INFO)


@pytest.fixture(scope="function")
def cube(com_config, dev_config, demo_opcua_server: DemoServer):
    cube = SupercubeBase(com_config, dev_config)
    cube.start()
    yield cube
    cube.stop()


def test_dev_config(dev_config):
    # currently there are no non-default config values
    SupercubeConfiguration()

    config = SupercubeConfiguration(**dev_config)
    for key, value in dev_config.items():
        assert getattr(config, key) == value


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"namespace_index": -1},
        {"polling_interval_sec": -1},
        {"polling_interval_sec": 0},
        {"polling_delay_sec": -1},
    ],
)
def test_invalid_config_dict(dev_config, wrong_config_dict):
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(ValueError):
        SupercubeConfiguration(**invalid_config)


def test_set_remote_control(cube: SupercubeBase, demo_opcua_server: DemoServer):
    cube.set_remote_control(False)
    assert demo_opcua_server.get_var(constants.OpcControl.active) is False
    cube.set_remote_control(True)
    assert demo_opcua_server.get_var(constants.OpcControl.active) is True


def test_status_poller(
    com_config, dev_config, demo_opcua_server: DemoServer, mocker: MockerFixture,
):

    cube = SupercubeBase(com_config, dev_config)
    poller = cube.status_poller
    assert poller is not None
    assert poller.polling_delay_sec == dev_config["polling_delay_sec"]
    assert poller.polling_interval_sec == dev_config["polling_interval_sec"]
    assert poller.polling_timeout_sec is None
    assert poller.spoll_handler == cube._spoll_handler

    spy_spoll_handler = mocker.spy(poller, "spoll_handler")
    spy_start_polling = mocker.spy(poller, "start_polling")
    spy_stop_polling = mocker.spy(poller, "stop_polling")

    poller.stop_polling()
    assert spy_stop_polling.spy_return is False

    cube.start()

    spy_start_polling.assert_called()
    assert spy_start_polling.spy_return is True

    poller.start_polling()
    assert spy_start_polling.spy_return is False

    sleep(dev_config["polling_delay_sec"] + 3 * dev_config["polling_interval_sec"])
    spy_spoll_handler.assert_called()
    assert spy_spoll_handler.spy_return is None

    cube.stop()
    spy_stop_polling.assert_called()
    assert spy_stop_polling.spy_return is True


def test_status_poller_timeout_error(
    com_config, dev_config, demo_opcua_server: DemoServer, mocker: MockerFixture,
):

    cube = SupercubeBase(com_config, dev_config)
    cube.start()
    assert cube.status_poller.is_polling()
    assert cube.com.is_open

    # patch UASocketClient.send_request to raise a mock TimeoutError as if coming from
    # used therein concurrent.futures.Future;
    # raise error only on disable remote control done via poller thread on stop()
    # => write(constants.OpcControl.active, False)

    # Use bound method (otherwise live unpatch does not work):
    send_request_orig = cube.com._client.uaclient._uasocket.send_request

    def send_request(self, request, *args, **kwargs):
        if (
            hasattr(request, "Parameters")
            and hasattr(request.Parameters, "NodesToWrite")
        ):
            node_to_write = request.Parameters.NodesToWrite[0]
            if (
                node_to_write.NodeId.Identifier == constants.OpcControl.active
                and node_to_write.Value.Value.Value is False
            ):
                from concurrent.futures import TimeoutError

                raise TimeoutError("mock timeout error")
        # method already bound - ignore `self`
        return send_request_orig(request, *args, **kwargs)

    mocker.patch(
        "opcua.client.ua_client.UASocketClient.send_request",
        side_effect=send_request,
        autospec=True,
    )

    # check poller thread error caught and wrapped
    with pytest.raises(OpcUaCommunicationTimeoutError):
        cube.set_remote_control(False)
    assert not cube.status_poller.is_polling()
    assert cube.com.is_open

    # check that stopping also does raises the error, but cleans up otherwise
    with pytest.raises(OpcUaCommunicationTimeoutError):
        cube.stop()
    assert not cube.status_poller.is_polling()
    assert not cube.com.is_open

    # unpatch and try to stop dev again - will raise error due to broken com,
    # but not the timeout
    mocker.patch(
        "opcua.client.ua_client.UASocketClient.send_request",
        side_effect=send_request_orig,
    )

    with pytest.raises(OpcUaCommunicationIOError) as excinfo:
        cube.stop()
        assert excinfo.type is not OpcUaCommunicationTimeoutError
    assert not cube.status_poller.is_polling()
    assert not cube.com.is_open


def test_get_support_input(cube: SupercubeBase):
    assert cube.get_support_input(1, 1) is False

    with pytest.raises(ValueError):
        cube.get_support_input(1, 123)
    with pytest.raises(ValueError):
        cube.get_support_input(123, 1)


def test_get_set_support_output(cube: SupercubeBase):
    cube.set_support_output(1, 1, False)
    assert cube.get_support_output(1, 1) is False
    cube.set_support_output(1, 1, True)
    assert cube.get_support_output(1, 1) is True

    with pytest.raises(ValueError):
        cube.get_support_output(1, 123)
    with pytest.raises(ValueError):
        cube.get_support_output(123, 1)
    with pytest.raises(ValueError):
        cube.set_support_output(1, 123, False)
    with pytest.raises(ValueError):
        cube.set_support_output(123, 1, False)


def test_set_support_output_impulse(cube: SupercubeBase):
    cube.set_support_output_impulse(2, 2, 0.01, True)

    with pytest.raises(ValueError):
        cube.set_support_output_impulse(1, 123)
    with pytest.raises(ValueError):
        cube.set_support_output_impulse(123, 1)


def test_get_t13_socket(cube: SupercubeBase):
    assert cube.get_t13_socket(1) is False
    cube.set_t13_socket(1, True)
    assert cube.get_t13_socket(1) is True

    with pytest.raises(ValueError):
        cube.get_t13_socket(4)


def test_set_t13_socket(cube: SupercubeBase):
    cube.set_t13_socket(1, True)

    with pytest.raises(ValueError):
        cube.set_t13_socket(4, False)

    with pytest.raises(ValueError):
        cube.set_t13_socket(1, "on")


def test_get_cee16(cube: SupercubeBase):
    cube.set_cee16_socket(False)
    assert cube.get_cee16_socket() is False
    cube.set_cee16_socket(True)
    assert cube.get_cee16_socket() is True


def test_set_cee16(cube: SupercubeBase):
    cube.set_cee16_socket(True)
    cube.set_cee16_socket(False)

    with pytest.raises(ValueError):
        cube.set_cee16_socket(1)

    with pytest.raises(ValueError):
        cube.set_cee16_socket("on")


def test_get_status(cube: SupercubeBase, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(constants.Safety.status, 1, False)
    assert cube.get_status() == constants.SafetyStatus.GreenNotReady

    demo_opcua_server.set_var(constants.Safety.status, 2)
    assert cube.get_status() == constants.SafetyStatus.GreenReady


def test_ready(cube: SupercubeBase, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(constants.Safety.switch_to_ready, False, True)
    cube.ready(True)
    assert demo_opcua_server.get_var(constants.Safety.switch_to_ready) is True


def test_operate(cube: SupercubeBase, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(constants.Safety.switch_to_operate, False, True)
    cube.operate(True)
    assert demo_opcua_server.get_var(constants.Safety.switch_to_operate) is True


def test_get_measurement_ratio(cube: SupercubeBase, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(constants.MeasurementsDividerRatio.input_1, 123.4, False)
    assert cube.get_measurement_ratio(1) == 123.4

    with pytest.raises(ValueError):
        cube.get_measurement_ratio(123)


def test_get_measurement_voltage(cube: SupercubeBase, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(constants.MeasurementsScaledInput.input_1, 110_000, False)
    assert cube.get_measurement_voltage(1) == 110_000.0

    with pytest.raises(ValueError):
        cube.get_measurement_voltage(123)


def test_get_earthing_stick_status(cube: SupercubeBase, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(constants.EarthingStick.status_1, 1, False)
    assert cube.get_earthing_stick_status(1) == constants.EarthingStickStatus(1)

    with pytest.raises(ValueError):
        cube.get_earthing_stick_status(123)


def test_get_earthing_stick_operating_status(
    cube: SupercubeBase, demo_opcua_server: DemoServer
):
    demo_opcua_server.add_var(
        constants.EarthingStick.operating_status_1,
        constants.EarthingStickOperatingStatus.auto,
        False,
    )
    assert (
        cube.get_earthing_stick_operating_status(1)
        is constants.EarthingStickOperatingStatus.auto
    )

    with pytest.raises(ValueError):
        cube.get_earthing_stick_operating_status(123)


def test_get_earthing_stick_manual(cube: SupercubeBase, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(
        constants.EarthingStick.manual_1,
        bool(constants.EarthingStickOperation.close),
        False,
    )
    assert cube.get_earthing_stick_manual(1) is constants.EarthingStickOperation.close

    with pytest.raises(ValueError):
        cube.get_earthing_stick_manual(123)


def test_operate_earthing_stick(cube: SupercubeBase, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(
        constants.EarthingStick.operating_status_1,
        constants.EarthingStickOperatingStatus.manual,
        False,
    )
    demo_opcua_server.add_var(
        constants.EarthingStick.manual_1,
        bool(constants.EarthingStickOperation.open),
        True,
    )
    assert cube.get_earthing_stick_manual(1) is constants.EarthingStickOperation.open
    cube.operate_earthing_stick(1, constants.EarthingStickOperation.close)
    assert cube.get_earthing_stick_manual(1) is constants.EarthingStickOperation.close


def test_operate_earthing_stick_error(
    cube: SupercubeBase, demo_opcua_server: DemoServer,
):
    demo_opcua_server.add_var(
        constants.EarthingStick.operating_status_1,
        constants.EarthingStickOperatingStatus.auto,
        False,
    )
    with pytest.raises(SupercubeEarthingStickOperationError):
        cube.operate_earthing_stick(1, constants.EarthingStickOperation.close)


def test_quit_error(cube: SupercubeBase, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(constants.Errors.quit, False, True)
    cube.quit_error()
    assert demo_opcua_server.get_var(constants.Errors.quit) is False


def test_get_door_status(cube: SupercubeBase, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(constants.Door.status_1, 2, False)
    assert cube.get_door_status(1) == constants.DoorStatus(2)


def test_get_earthing_rod_status(cube: SupercubeBase, demo_opcua_server: DemoServer):
    demo_opcua_server.add_var(constants.EarthingRod.status_1, 0, False)
    assert cube.get_earthing_rod_status(1) == constants.EarthingRodStatus(0)


def test_set_status_board(cube: SupercubeBase, demo_opcua_server: DemoServer):

    n_max = len(constants.MessageBoard)

    with pytest.raises(ValueError):
        cube.set_status_board(["x"] * (n_max + 1))
    with pytest.raises(ValueError):
        cube.set_status_board(["x"], pos=[n_max])

    for line in constants.MessageBoard:
        demo_opcua_server.add_var(line, "", True)

    msgs = ["Hello World", "Hello Fabian", "Hello HVL"]

    cube.set_status_board(msgs[:1])
    assert demo_opcua_server.get_var(constants.MessageBoard.line(1)) == msgs[0]
    cube.set_status_board(msgs[:1], pos=[1], clear_board=False)
    assert demo_opcua_server.get_var(constants.MessageBoard.line(1)) == msgs[0]
    assert demo_opcua_server.get_var(constants.MessageBoard.line(2)) == msgs[0]

    positions = [4, 8, 12]
    cube.set_status_board(msgs, positions, clear_board=False)
    for i, p in enumerate(positions):
        assert demo_opcua_server.get_var(constants.MessageBoard.line(p+1)) == msgs[i]

    cube.set_status_board([str(i) for i in range(15)])
    for i in range(15):
        assert demo_opcua_server.get_var(constants.MessageBoard.line(i+1)) == str(i)


def test_set_message_board(cube: SupercubeBase, demo_opcua_server: DemoServer):

    n_max = len(constants.MessageBoard)

    with pytest.raises(ValueError):
        cube.set_message_board(["x"] * (n_max + 1))

    for line in constants.MessageBoard:
        demo_opcua_server.add_var(line, "", True)

    msgs = ["ERROR: This is unexpected", "Good choice"]
    cube.set_message_board(msgs[:1])
    assert demo_opcua_server.get_var(constants.MessageBoard.line(1)).endswith(msgs[0])

    # push two messages to first two lines, in the given order
    cube.set_message_board(msgs)
    assert demo_opcua_server.get_var(constants.MessageBoard.line(1)).endswith(msgs[0])
    assert demo_opcua_server.get_var(constants.MessageBoard.line(2)).endswith(msgs[1])
    assert demo_opcua_server.get_var(constants.MessageBoard.line(3)).endswith(msgs[0])

    # overwrite all messages except last, which is pushed to be last line
    cube.set_message_board("x" * (n_max - 1))
    for n in range(len(constants.MessageBoard)-1):
        assert demo_opcua_server.get_var(constants.MessageBoard.line(n+1)).endswith("x")
    last_msg = demo_opcua_server.get_var(constants.MessageBoard.line(n_max))
    assert last_msg.endswith(msgs[0])



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Constants, variable names for the Supercube OPC-connected devices.
"""

import logging
from aenum import IntEnum

from hvl_ccb.utils.enum import ValueEnum, unique

logger = logging.getLogger(__name__)


class SupercubeOpcEndpoint(ValueEnum):
    """
    OPC Server Endpoint strings for the supercube variants.
    """

    A = "OPC.SimaticNET.S7"
    B = "OPC.SimaticNET.S7"


@unique
class GeneralSupport(ValueEnum):
    """
    NodeID strings for the support inputs and outputs.
    """

    in_1_1 = "hvl-ipc.WINAC.Support1InA"
    in_1_2 = "hvl-ipc.WINAC.Support1InB"
    in_2_1 = "hvl-ipc.WINAC.Support2InA"
    in_2_2 = "hvl-ipc.WINAC.Support2InB"
    in_3_1 = "hvl-ipc.WINAC.Support3InA"
    in_3_2 = "hvl-ipc.WINAC.Support3InB"
    in_4_1 = "hvl-ipc.WINAC.Support4InA"
    in_4_2 = "hvl-ipc.WINAC.Support4InB"
    in_5_1 = "hvl-ipc.WINAC.Support5InA"
    in_5_2 = "hvl-ipc.WINAC.Support5InB"
    in_6_1 = "hvl-ipc.WINAC.Support6InA"
    in_6_2 = "hvl-ipc.WINAC.Support6InB"
    out_1_1 = "hvl-ipc.WINAC.Support1OutA"
    out_1_2 = "hvl-ipc.WINAC.Support1OutB"
    out_2_1 = "hvl-ipc.WINAC.Support2OutA"
    out_2_2 = "hvl-ipc.WINAC.Support2OutB"
    out_3_1 = "hvl-ipc.WINAC.Support3OutA"
    out_3_2 = "hvl-ipc.WINAC.Support3OutB"
    out_4_1 = "hvl-ipc.WINAC.Support4OutA"
    out_4_2 = "hvl-ipc.WINAC.Support4OutB"
    out_5_1 = "hvl-ipc.WINAC.Support5OutA"
    out_5_2 = "hvl-ipc.WINAC.Support5OutB"
    out_6_1 = "hvl-ipc.WINAC.Support6OutA"
    out_6_2 = "hvl-ipc.WINAC.Support6OutB"

    @classmethod
    def output(cls, port, contact):
        """
        Get the NodeID string for a support output.

        :param port: the desired port (1..6)
        :param contact: the desired contact at the port (1..2)
        :return: the node id string
        """

        return getattr(cls, f"out_{port}_{contact}")

    @classmethod
    def input(cls, port, contact):
        """
        Get the NodeID string for a support input.

        :param port: the desired port (1..6)
        :param contact: the desired contact at the port (1..2)
        :return: the node id string
        """

        return getattr(cls, f"in_{port}_{contact}")


@unique
class BreakdownDetection(ValueEnum):
    """
    Node ID strings for the breakdown detection.
    """

    #: Boolean read-only variable indicating whether breakdown detection and fast
    #: switchoff is enabled in the system or not.
    activated = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.Breakdowndetection.connect"

    #: Boolean read-only variable telling whether the fast switch-off has triggered.
    #: This can also be seen using the safety circuit state, therefore no method is
    #: implemented to read this out directly.
    triggered = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.Breakdowndetection.triggered"

    #: Boolean writable variable to reset the fast switch-off. Toggle to re-enable.
    reset = "hvl-ipc.WINAC.Support6OutA"


@unique
class GeneralSockets(ValueEnum):
    """
    NodeID strings for the power sockets (3x T13 and 1xCEE16).
    """

    #: SEV T13 socket No. 1 (writable boolean).
    t13_1 = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.T13_1"

    #: SEV T13 socket No. 2 (writable boolean).
    t13_2 = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.T13_2"

    #: SEV T13 socket No. 3 (writable boolean).
    t13_3 = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.T13_3"

    #: CEE16 socket (writeable boolean).
    cee16 = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.CEE16"


T13_SOCKET_PORTS = (1, 2, 3)
"""
Port numbers of SEV T13 power socket
"""


@unique
class Safety(ValueEnum):
    """
    NodeID strings for the basic safety circuit status and green/red switches "ready"
    and "operate".
    """

    #: Status is a read-only integer containing the state number of the
    #: supercube-internal state machine. The values correspond to numbers in
    #: :class:`SafetyStatus`.
    status_ready_for_red = "hvl-ipc.WINAC.SYSTEMSTATE.ReadyForRed"
    status_red = "hvl-ipc.WINAC.SYSTEMSTATE.RED"
    status_green = "hvl-ipc.WINAC.SYSTEMSTATE.GREEN"
    status_error = "hvl-ipc.WINAC.SYSTEMSTATE.ERROR"

    #: Writable boolean for switching to Red Ready (locked, HV off) state.
    switchto_ready = "hvl-ipc.WINAC.SYSTEMSTATE.RED_REQUEST"
    switchto_green = "hvl-ipc.WINAC.SYSTEMSTATE.GREEN_REQUEST"

    #: Writable boolean for switching to Red Operate (locket, HV on) state.
    switchto_operate = "hvl-ipc.WINAC.SYSTEMSTATE.switchon"

    #: Writeable boolean to manually turn on or off the horn
    horn = "hvl-ipc.WINAC.SYSTEM_INTERN.hornen"


class SafetyStatus(IntEnum):
    """
    Safety status values that are possible states returned from
    :meth:`hvl_ccb.dev.supercube.base.Supercube.get_status`. These
    values correspond to the states of the Supercube's safety circuit statemachine.
    """

    #: System is initializing or booting.
    Initializing = 0

    #: System is safe, lamps are green and some safety elements are not in place such
    #: that it cannot be switched to red currently.
    GreenNotReady = 1

    #: System is safe and all safety elements are in place to be able to switch to
    #: *ready*.
    GreenReady = 2

    #: System is locked in red state and *ready* to go to *operate* mode.
    RedReady = 3

    #: System is locked in red state and in *operate* mode, i.e. high voltage on.
    RedOperate = 4

    #: Fast turn off triggered and switched off the system. Reset FSO to go back to a
    #: normal state.
    QuickStop = 5

    #: System is in error mode.
    Error = 6


@unique
class Power(ValueEnum):
    """
    Variable NodeID strings concerning power data.
    """

    #: Primary voltage in volts, measured by the frequency converter at its output.
    #: (read-only)
    voltage_primary = "hvl-ipc.WINAC.SYSTEM_INTERN.FUVoltageprim"

    #: Primary current in ampere, measured by the frequency converter. (read-only)
    current_primary = "hvl-ipc.WINAC.SYSTEM_INTERN.FUCurrentprim"

    #: Power setup that is configured using the Supercube HMI. The value corresponds to
    #: the ones in :class:`PowerSetup`. (read-only)
    setup = "hvl-ipc.WINAC.FU.TrafoSetup"

    #: Voltage slope in V/s.
    voltage_slope = "hvl-ipc.WINAC.FU.dUdt_-1"

    #: Target voltage setpoint in V.
    voltage_target = "hvl-ipc.WINAC.FU.SOLL"

    #: Maximum voltage allowed by the current experimental setup. (read-only)
    voltage_max = "hvl-ipc.WINAC.FU.maxVoltagekV"

    #: Frequency converter output frequency. (read-only)
    frequency = "hvl-ipc.WINAC.FU.Frequency"


class PowerSetup(IntEnum):
    """
    Possible power setups corresponding to the value of variable :attr:`Power.setup`.
    """

    #: No safety switches, use only safety components (doors, fence, earthing...)
    #: without any power.
    # NoPower = 0

    #: External power supply fed through blue CEE32 input using isolation transformer
    #: and safety switches of the Supercube, or using an external safety switch
    #: attached to the Supercube Type B.
    External = 0

    #: AC voltage with MWB transformer set to 50kV maximum voltage.
    AC_SingleStage_50kV = 1

    #: AC voltage with MWB transformer set to 100kV maximum voltage.
    AC_SingleStage_100kV = 2

    #: AC voltage with two MWB transformers, one at 100kV and the other at 50kV,
    #: resulting in a total maximum voltage of 150kV.
    AC_DoubleStage_150kV = 3

    #: AC voltage with two MWB transformers both at 100kV, resulting in a total
    #: maximum voltage of 200kV
    AC_DoubleStage_200kV = 4

    #: Internal usage of the frequency converter, controlling to the primary voltage
    #: output of the supercube itself (no measurement transformer used)
    Internal = 5

    #: DC voltage with one AC transformer set to 100kV AC, resulting in 140kV DC
    DC_SingleStage_140kV = 6

    #: DC voltage with two AC transformers set to 100kV AC each, resulting in 280kV
    #: DC in total (or a single stage transformer with Greinacher voltage doubling
    #: rectifier)
    DC_DoubleStage_280kV = 7


@unique
class MeasurementsScaledInput(ValueEnum):
    """
    Variable NodeID strings for the four analog BNC inputs for measuring voltage.
    The voltage returned in these variables is already scaled with the set ratio,
    which can be read using the variables in :class:`MeasurementsDividerRatio`.
    """

    input_1 = "hvl-ipc.WINAC.SYSTEM_INTERN.AI1Volt"
    input_2 = "hvl-ipc.WINAC.SYSTEM_INTERN.AI2Volt"
    input_3 = "hvl-ipc.WINAC.SYSTEM_INTERN.AI3Volt"
    input_4 = "hvl-ipc.WINAC.SYSTEM_INTERN.AI4Volt"

    @classmethod
    def get(cls, channel: int):
        """
        Get the attribute for an input number.

        :param channel: the channel number (1..4)
        :return: the enum for the desired channel.
        """

        return getattr(cls, f"input_{channel}")


@unique
class MeasurementsDividerRatio(ValueEnum):
    """
    Variable NodeID strings for the measurement input scaling ratios. These ratios
    are defined in the Supercube HMI setup and are provided in the python module here
    to be able to read them out, allowing further calculations.
    """

    input_1 = "hvl-ipc.WINAC.SYSTEM_INTERN.DivididerRatio"

    @classmethod
    def get(cls, channel: int):
        """
        Get the attribute for an input number.

        :param channel: the channel number (1..4)
        :return: the enum for the desired channel.
        """

        return getattr(cls, f"input_{channel}")


class EarthingStickStatus(IntEnum):
    """
    Status of an earthing stick. These are the possible values in the status integer
    e.g. in :attr:`EarthingStick.status_1`.
    """

    #: Earthing stick is deselected and not enabled in safety circuit. To get out of
    #: this state, the earthing has to be enabled in the Supercube HMI setup.
    inactive = 0

    #: Earthing is closed (safe).
    closed = 1

    #: Earthing is open (not safe).
    open = 2

    #: Earthing is in error, e.g. when the stick did not close correctly or could not
    #: open.
    error = 3


@unique
class EarthingStick(ValueEnum):
    """
    Variable NodeID strings for all earthing stick statuses (read-only integer) and
    writable booleans for setting the earthing in manual mode.
    """

    status_1_closed = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_1.CLOSE"
    status_1_open = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_1.OPEN"
    status_1_connected = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_1.CONNECT"
    status_2_closed = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_2.CLOSE"
    status_2_open = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_2.OPEN"
    status_2_connected = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_2.CONNECT"
    status_3_closed = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_3.CLOSE"
    status_3_open = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_3.OPEN"
    status_3_connected = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_3.CONNECT"
    status_4_closed = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_4.CLOSE"
    status_4_open = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_4.OPEN"
    status_4_connected = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_4.CONNECT"
    status_5_closed = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_5.CLOSE"
    status_5_open = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_5.OPEN"
    status_5_connected = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_5.CONNECT"
    status_6_closed = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_6.CLOSE"
    status_6_open = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_6.OPEN"
    status_6_connected = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_6.CONNECT"

    manual_1 = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_1.MANUAL"
    manual_2 = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_2.MANUAL"
    manual_3 = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_3.MANUAL"
    manual_4 = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_4.MANUAL"
    manual_5 = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_5.MANUAL"
    manual_6 = "hvl-ipc.WINAC.SYSTEM_COMPONENTS.STICK_6.MANUAL"

    @classmethod
    def status_open(cls, number: int):
        """
        Get the status enum attribute for an earthing stick number.

        :param number: the earthing stick (1..6)
        :return: the status enum
        """

        return getattr(cls, f"status_{number}_open")

    @classmethod
    def status_closed(cls, number: int):
        """
        Get the status enum attribute for an earthing stick number.

        :param number: the earthing stick (1..6)
        :return: the status enum
        """

        return getattr(cls, f"status_{number}_closed")

    @classmethod
    def status_connected(cls, number: int):
        """
        Get the status enum attribute for an earthing stick number.

        :param number: the earthing stick (1..6)
        :return: the status enum
        """

        return getattr(cls, f"status_{number}_connected")

    @classmethod
    def manual(cls, number: int):
        """
        Get the manual enum attribute for an earthing stick number.

        :param number: the earthing stick (1..6)
        :return: the manual enum
        """

        return getattr(cls, f"manual_{number}")


@unique
class Errors(ValueEnum):
    """
    Variable NodeID strings for information regarding error, warning and message
    handling.
    """

    #: Boolean read-only variable telling if a stop is active.
    stop = "hvl-ipc.WINAC.SYSTEMSTATE.ERROR"
    stop_number = "hvl-ipc.WINAC.SYSTEMSTATE.Errornumber"

    #: Writable boolean for the error quit button.
    quit = "hvl-ipc.WINAC.SYSTEMSTATE.Faultconfirmation"


class DoorStatus(IntEnum):
    """
    Possible status values for doors.
    """

    #: not enabled in Supercube HMI setup, this door is not supervised.
    inactive = 0

    #: Door is open.
    open = 1

    #: Door is closed, but not locked.
    closed = 2

    #: Door is closed and locked (safe state).
    locked = 3

    #: Door has an error or was opened in locked state (either with emergency stop or
    #: from the inside).
    error = 4


class AlarmText(ValueEnum):
    """
    This enumeration contains textual representations for all error classes (stop,
    warning and message) of the Supercube system. Use the :meth:`AlarmText.get`
    method to retrieve the enum of an alarm number.
    """

    # No alarm
    Alarm0 = "No Alarm."

    # Safety elements
    Alarm1 = "STOP Safety switch 1 error"
    Alarm2 = "STOP Safety switch 2 error"
    Alarm3 = "STOP Emergency Stop 1"
    Alarm4 = "STOP Emergency Stop 2"
    Alarm5 = "STOP Emergency Stop 3"
    Alarm6 = "STOP Door 1 lock supervision"
    Alarm7 = "STOP Door 2 lock supervision"
    Alarm8 = "STOP Door 3 lock supervision"
    Alarm9 = "STOP Earthing stick 1 error"
    Alarm10 = "STOP Earthing stick 2 error"
    Alarm11 = "STOP Earthing stick 3 error"
    Alarm12 = "STOP Earthing stick 4 error"
    Alarm13 = "STOP Earthing stick 5 error"
    Alarm14 = "STOP Earthing stick 6 error"
    Alarm17 = "STOP Source switch error"
    Alarm19 = "STOP Fence 1 error"
    Alarm20 = "STOP Fence 2 error"
    Alarm21 = "STOP Control error"
    Alarm22 = "STOP Power outage"

    # generic not defined alarm text
    not_defined = "NO ALARM TEXT DEFINED"

    @classmethod
    def get(cls, alarm: int):
        """
        Get the attribute of this enum for an alarm number.

        :param alarm: the alarm number
        :return: the enum for the desired alarm number
        """

        try:
            return getattr(cls, f"Alarm{alarm}")
        except AttributeError as e:
            logger.error(str(e), exc_info=e)
            return cls.not_defined



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Test module for the supercube constants enums.
"""

import pytest

from hvl_ccb.dev.supercube import constants


def test_support_output():
    assert constants.GeneralSupport.output(1, 1) == constants.GeneralSupport.out_1_1
    assert constants.GeneralSupport.output(5, 2) == constants.GeneralSupport.out_5_2
    with pytest.raises(ValueError):
        constants.GeneralSupport.output(3, 3)


def test_support_input():
    assert constants.GeneralSupport.input(1, 1) == constants.GeneralSupport.in_1_1
    assert constants.GeneralSupport.input(5, 2) == constants.GeneralSupport.in_5_2
    with pytest.raises(ValueError):
        constants.GeneralSupport.input(3, 3)


def test_measurements_scaled_input():
    assert (constants.MeasurementsScaledInput.input(1) ==
            constants.MeasurementsScaledInput.input_1)
    assert (constants.MeasurementsScaledInput.input(3) ==
            constants.MeasurementsScaledInput.input_3)
    for n in (0, 5):
        with pytest.raises(ValueError):
            constants.MeasurementsScaledInput.input(n)


def test_measurements_divider_ratio():
    assert (constants.MeasurementsDividerRatio.input(1) ==
            constants.MeasurementsDividerRatio.input_1)
    assert (constants.MeasurementsDividerRatio.input(3) ==
            constants.MeasurementsDividerRatio.input_3)
    for n in (0, 5):
        with pytest.raises(ValueError):
            constants.MeasurementsDividerRatio.input(n)


def test_earthing_stick_status():
    assert (constants.EarthingStick.status(3) ==
            constants.EarthingStick.status_3)
    with pytest.raises(AttributeError):
        constants.MeasurementsDividerRatio.status(7)


def test_earthing_stick_manual():
    assert (constants.EarthingStick.manual(2) ==
            constants.EarthingStick.manual_2)
    with pytest.raises(ValueError):
        constants.EarthingStick.manual(7)


def test_alarms():
    assert constants.Alarms.Alarm(33) == constants.Alarms.Alarm33
    for n in (0, 152):
        with pytest.raises(ValueError):
            constants.Alarms.Alarm(n)


def test_door():
    assert constants.Door.status(3) == constants.Door.status_3
    for n in (0, 4):
        with pytest.raises(ValueError):
            constants.Door.status(n)


def test_alarm_text():
    assert constants.AlarmText.get(1) == constants.AlarmText.Alarm1
    assert constants.AlarmText.get(1000) == constants.AlarmText.not_defined


def test_message_board():
    assert constants.MessageBoard.line(3) == constants.MessageBoard.line_3
    for n in (0, 16):
        with pytest.raises(ValueError):
            constants.MessageBoard.line(n)



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Base classes for the Supercube device.
"""

import logging
from collections import deque
from datetime import datetime
from itertools import cycle
from time import sleep
from typing import List, Sequence

from opcua import Node

from . import constants
from ..base import SingleCommDevice
from ..utils import Poller
from ...comm import (
    OpcUaCommunication,
    OpcUaCommunicationConfig,
    OpcUaSubHandler,
)
from ...configuration import configdataclass
from ...utils.typing import Number

logger = logging.getLogger(__name__)


class SupercubeEarthingStickOperationError(Exception):
    pass


class SupercubeSubscriptionHandler(OpcUaSubHandler):
    """
    OPC Subscription handler for datachange events and normal events specifically
    implemented for the Supercube devices.
    """

    def datachange_notification(self, node: Node, val, data):
        """
        In addition to the standard operation (debug logging entry of the datachange),
        alarms are logged at INFO level using the alarm text.

        :param node: the node object that triggered the datachange event
        :param val: the new value
        :param data:
        """

        super().datachange_notification(node, val, data)

        # assume an alarm datachange
        try:
            alarm_number = constants.Alarms(node.nodeid.Identifier).number
            alarm_text = constants.AlarmText.get(alarm_number)
            alarm_log_prefix = "Coming" if val else "Going"
            logger.info(f"{alarm_log_prefix} Alarm {alarm_number}: {str(alarm_text)}")
            return
        except ValueError:
            # not any of Alarms node IDs
            pass

        id_ = node.nodeid.Identifier

        # assume a status datachange
        if id_ == str(constants.Safety.status):
            new_status = str(constants.SafetyStatus(val))
            logger.info(f"Safety: {new_status}")
            return

        # assume an earthing stick status datachange
        if id_ in constants.EarthingStick.statuses():
            new_status = str(constants.EarthingStickStatus(val))
            logger.info(f"Earthing {constants.EarthingStick(id_).number}: {new_status}")
            return


@configdataclass
class SupercubeConfiguration:
    """
    Configuration dataclass for the Supercube devices.
    """

    #: Namespace of the OPC variables, typically this is 3 (coming from Siemens)
    namespace_index: int = 3

    polling_delay_sec: Number = 5.0
    polling_interval_sec: Number = 1.0

    def clean_values(self):
        if self.namespace_index < 0:
            raise ValueError(
                "Index of the OPC variables namespace needs to be a positive integer."
            )
        if self.polling_interval_sec <= 0:
            raise ValueError("Polling interval needs to be positive.")
        if self.polling_delay_sec < 0:
            raise ValueError("Polling delay needs to be not negative.")


@configdataclass
class SupercubeOpcUaCommunicationConfig(OpcUaCommunicationConfig):
    """
    Communication protocol configuration for OPC UA, specifications for the Supercube
    devices.
    """

    #: Subscription handler for data change events
    sub_handler: OpcUaSubHandler = SupercubeSubscriptionHandler()


class SupercubeOpcUaCommunication(OpcUaCommunication):
    """
    Communication protocol specification for Supercube devices.
    """

    @staticmethod
    def config_cls():
        return SupercubeOpcUaCommunicationConfig


class SupercubeBase(SingleCommDevice):
    """
    Base class for Supercube variants.
    """

    def __init__(self, com, dev_config=None):
        """
        Constructor for Supercube base class.

        :param com: the communication protocol or its configuration
        :param dev_config: the device configuration
        """

        super().__init__(com, dev_config)

        self.status_poller = Poller(
            self._spoll_handler,
            polling_delay_sec=self.config.polling_delay_sec,
            polling_interval_sec=self.config.polling_interval_sec,
        )
        self.toggle = cycle([False, True])
        self.message_len = len(constants.MessageBoard)
        self.status_board = [""] * self.message_len
        self.message_board = deque([""] * self.message_len, maxlen=self.message_len)

    @staticmethod
    def default_com_cls():
        return SupercubeOpcUaCommunication

    def start(self) -> None:
        """
        Starts the device. Sets the root node for all OPC read and write commands to
        the Siemens PLC object node which holds all our relevant objects and variables.
        """

        logger.info("Starting Supercube Base device")
        super().start()

        logger.debug("Add monitoring nodes")
        self.com.init_monitored_nodes(
            map(  # type: ignore
                str, constants.GeneralSockets
            ),
            self.config.namespace_index,
        )
        self.com.init_monitored_nodes(
            map(  # type: ignore
                str, constants.GeneralSupport
            ),
            self.config.namespace_index,
        )
        self.com.init_monitored_nodes(
            map(  # type: ignore
                str, constants.Safety
            ),
            self.config.namespace_index,
        )
        self.com.init_monitored_nodes(
            str(constants.Errors.message), self.config.namespace_index
        )
        self.com.init_monitored_nodes(
            str(constants.Errors.warning), self.config.namespace_index
        )
        self.com.init_monitored_nodes(
            str(constants.Errors.stop), self.config.namespace_index
        )
        self.com.init_monitored_nodes(
            map(str, constants.Alarms), self.config.namespace_index
        )
        self.com.init_monitored_nodes(
            map(  # type: ignore
                str, constants.EarthingStick
            ),
            self.config.namespace_index,
        )

        self.set_remote_control(True)
        logger.debug("Finished starting")

    def stop(self) -> None:
        """
        Stop the Supercube device. Deactivates the remote control and closes the
        communication protocol.
        """
        try:
            self.set_remote_control(False)
        finally:
            super().stop()

    def _spoll_handler(self) -> None:
        """
        Supercube poller handler; change one byte on a SuperCube.
        """
        self.write(constants.OpcControl.live, next(self.toggle))

    @staticmethod
    def config_cls():
        return SupercubeConfiguration

    def read(self, node_id: str):
        """
        Local wrapper for the OPC UA communication protocol read method.

        :param node_id: the id of the node to read.
        :return: the value of the variable
        """

        logger.debug(f"Read from node ID {node_id} ...")
        result = self.com.read(str(node_id), self.config.namespace_index)
        logger.debug(f"Read from node ID {node_id}: {result}")
        return result

    def write(self, node_id, value) -> None:
        """
        Local wrapper for the OPC UA communication protocol write method.

        :param node_id: the id of the node to read
        :param value: the value to write to the variable
        """
        logger.debug(f"Write to node ID {node_id}: {value}")
        self.com.write(str(node_id), self.config.namespace_index, value)

    def set_remote_control(self, state: bool) -> None:
        """
        Enable or disable remote control for the Supercube. This will effectively
        display a message on the touchscreen HMI.

        :param state: desired remote control state
        """
        can_write = False
        try:
            self.write(constants.OpcControl.active, bool(state))
            can_write = True
        finally:
            if state:
                if not can_write:
                    logger.warning("Remote control cannot be enabled")
                else:
                    was_not_polling = self.status_poller.start_polling()
                    if not was_not_polling:
                        logger.warning("Remote control already enabled")
            else:
                was_polling = self.status_poller.stop_polling()
                self.status_poller.wait_for_polling_result()
                if not was_polling:
                    logger.warning("Remote control already disabled")

    def get_support_input(self, port: int, contact: int) -> bool:
        """
        Get the state of a support socket input.

        :param port: is the socket number (1..6)
        :param contact: is the contact on the socket (1..2)
        :return: digital input read state
        :raises ValueError: when port or contact number is not valid
        """

        return bool(self.read(constants.GeneralSupport.input(port, contact)))

    def get_support_output(self, port: int, contact: int) -> bool:
        """
        Get the state of a support socket output.

        :param port: is the socket number (1..6)
        :param contact: is the contact on the socket (1..2)
        :return: digital output read state
        :raises ValueError: when port or contact number is not valid
        """

        return bool(self.read(constants.GeneralSupport.output(port, contact)))

    def set_support_output(self, port: int, contact: int, state: bool) -> None:
        """
        Set the state of a support output socket.

        :param port: is the socket number (1..6)
        :param contact: is the contact on the socket (1..2)
        :param state: is the desired state of the support output
        :raises ValueError: when port or contact number is not valid
        """

        self.write(constants.GeneralSupport.output(port, contact), bool(state))

    def set_support_output_impulse(
        self, port: int, contact: int, duration: float = 0.2, pos_pulse: bool = True
    ) -> None:
        """
        Issue an impulse of a certain duration on a support output contact. The polarity
        of the pulse (On-wait-Off or Off-wait-On) is specified by the pos_pulse
        argument.

        This function is blocking.

        :param port: is the socket number (1..6)
        :param contact: is the contact on the socket (1..2)
        :param duration: is the length of the impulse in seconds
        :param pos_pulse: is True, if the pulse shall be HIGH, False if it shall be LOW
        :raises ValueError: when port or contact number is not valid
        """

        self.set_support_output(port, contact, pos_pulse)
        sleep(duration)
        self.set_support_output(port, contact, not pos_pulse)

    def get_t13_socket(self, port: int) -> bool:
        """
        Read the state of a SEV T13 power socket.

        :param port: is the socket number, one of `constants.T13_SOCKET_PORTS`
        :return: on-state of the power socket
        :raises ValueError: when port is not valid
        """

        if port not in constants.T13_SOCKET_PORTS:
            raise ValueError(f"port not in {constants.T13_SOCKET_PORTS}: {port}")

        return bool(self.read(getattr(constants.GeneralSockets, f"t13_{port}")))

    def set_t13_socket(self, port: int, state: bool) -> None:
        """
        Set the state of a SEV T13 power socket.

        :param port: is the socket number, one of `constants.T13_SOCKET_PORTS`
        :param state: is the desired on-state of the socket
        :raises ValueError: when port is not valid or state is not of type bool
        """

        if not isinstance(state, bool):
            raise ValueError(f"state is not <bool>: {state}")

        if port not in constants.T13_SOCKET_PORTS:
            raise ValueError(f"port not in {constants.T13_SOCKET_PORTS}: {port}")

        self.write(getattr(constants.GeneralSockets, f"t13_{port}"), state)

    def get_cee16_socket(self) -> bool:
        """
        Read the on-state of the IEC CEE16 three-phase power socket.

        :return: the on-state of the CEE16 power socket
        """

        return bool(self.read(constants.GeneralSockets.cee16))

    def set_cee16_socket(self, state: bool) -> None:
        """
        Switch the IEC CEE16 three-phase power socket on or off.

        :param state: desired on-state of the power socket
        :raises ValueError: if state is not of type bool
        """

        if not isinstance(state, bool):
            raise ValueError(f"state is not <bool>: {state}")

        self.write(constants.GeneralSockets.cee16, state)

    def get_status(self) -> constants.SafetyStatus:
        """
        Get the safety circuit status of the Supercube.
        :return: the safety status of the supercube's state machine.
        """

        return constants.SafetyStatus(self.read(constants.Safety.status))

    def ready(self, state: bool) -> None:
        """
        Set ready state. Ready means locket safety circuit, red lamps, but high voltage
        still off.

        :param state: set ready state
        """

        self.write(constants.Safety.switch_to_ready, state)

    def operate(self, state: bool) -> None:
        """
        Set operate state. If the state is RedReady, this will turn on the high
        voltage and close the safety switches.

        :param state: set operate state
        """

        self.write(constants.Safety.switch_to_operate, state)

    def get_measurement_ratio(self, channel: int) -> float:
        """
        Get the set measurement ratio of an AC/DC analog input channel. Every input
        channel has a divider ratio assigned during setup of the Supercube system.
        This ratio can be read out.

        :param channel: number of the input channel (1..4)
        :return: the ratio
        :raises ValueError: when channel is not valid
        """

        return float(self.read(constants.MeasurementsDividerRatio.input(channel)))

    def get_measurement_voltage(self, channel: int) -> float:
        """
        Get the measured voltage of an analog input channel. The voltage read out
        here is already scaled by the configured divider ratio.

        :param channel: number of the input channel (1..4)
        :return: measured voltage
        :raises ValueError: when channel is not valid
        """

        return float(self.read(constants.MeasurementsScaledInput.input(channel)))

    def get_earthing_stick_status(self, number: int) -> constants.EarthingStickStatus:
        """
        Get the status of an earthing stick, whether it is closed, open or undefined
        (moving).

        :param number: number of the earthing stick (1..6)
        :return: earthing stick status
        :raises ValueError: when earthing stick number is not valid
        """

        return constants.EarthingStickStatus(
            self.read(constants.EarthingStick.status(number))
        )

    def get_earthing_stick_operating_status(
        self, number: int
    ) -> constants.EarthingStickOperatingStatus:
        """
        Get the operating status of an earthing stick.

        :param number: number of the earthing stick (1..6)
        :return: earthing stick operating status (auto == 0, manual == 1)
        :raises ValueError: when earthing stick number is not valid
        """
        return constants.EarthingStickOperatingStatus(
            self.read(constants.EarthingStick.operating_status(number))
        )

    def get_earthing_stick_manual(
        self, number: int
    ) -> constants.EarthingStickOperation:
        """
        Get the manual status of an earthing stick. If an earthing stick is set to
        manual, it is closed even if the system is in states RedReady or RedOperate.

        :param number: number of the earthing stick (1..6)
        :return: operation of the earthing stick in a manual operating mode (open == 0,
            close == 1)
        :raises ValueError: when earthing stick number is not valid
        """
        return constants.EarthingStickOperation(
            int(self.read(constants.EarthingStick.manual(number)))
        )

    def operate_earthing_stick(
        self, number: int, operation: constants.EarthingStickOperation
    ) -> None:
        """
        Operation of an earthing stick, which is set to manual operation. If an earthing
        stick is set to manual, it stays closed even if the system is in states
        RedReady or RedOperate.

        :param number: number of the earthing stick (1..6)
        :param operation: earthing stick manual status (close or open)
        :raises SupercubeEarthingStickOperationError: when operating status of given
            number's earthing stick is not manual
        """

        if (
            self.get_earthing_stick_operating_status(number)
            == constants.EarthingStickOperatingStatus.manual
        ):
            self.write(constants.EarthingStick.manual(number), bool(operation.value))
        else:
            raise SupercubeEarthingStickOperationError

    def quit_error(self) -> None:
        """
        Quits errors that are active on the Supercube.
        """

        self.write(constants.Errors.quit, True)
        sleep(0.1)
        self.write(constants.Errors.quit, False)

    def get_door_status(self, door: int) -> constants.DoorStatus:
        """
        Get the status of a safety fence door. See :class:`constants.DoorStatus` for
        possible returned door statuses.

        :param door: the door number (1..3)
        :return: the door status
        """

        return constants.DoorStatus(self.read(constants.Door.status(door)))

    def get_earthing_rod_status(self, earthing_rod: int) -> constants.EarthingRodStatus:
        """
        Get the status of a earthing rod. See :class:`constants.EarthingRodStatus` for
        possible returned earthing rod statuses.

        :param earthing_rod: the earthing rod number (1..3)
        :return: the earthing rod status
        """
        value = self.read(constants.EarthingRod.status(earthing_rod))
        return constants.EarthingRodStatus(value)

    def set_status_board(
        self,
        msgs: List[str],
        pos: List[int] = None,
        clear_board: bool = True,
        display_board: bool = True,
    ) -> None:
        """
        Sets and displays a status board. The messages and the position of the message
        can be defined.

        :param msgs: list of strings
        :param pos: list of integers [0...14]
        :param clear_board: clear unspecified lines if `True` (default), keep otherwise
        :param display_board: display new status board if `True` (default)
        :raises ValueError: if there are too many messages or the positions indices are
            invalid.
        """
        # validate inputs
        if len(msgs) > self.message_len:
            raise ValueError(
                f"Too many message: {len(msgs)} given, max. {self.message_len} allowed."
            )
        if pos and not all(0 < p < self.message_len for p in pos):
            raise ValueError(f"Messages positions out of 0...{self.message_len} range")

        if clear_board:
            self.status_board = [""] * self.message_len

        # update status board
        if not pos:
            pos = list(range(len(msgs)))
        for num, msg in zip(pos, msgs):
            self.status_board[num] = msg
        if display_board:
            self.display_status_board()

    def display_status_board(self) -> None:
        """
        Display status board.
        """
        return self._display_messages(self.status_board)

    def set_message_board(self, msgs: List[str], display_board: bool = True) -> None:
        """
        Fills messages into message board that display that 15 newest messages with
        a timestamp.

        :param msgs: list of strings
        :param display_board: display 15 newest messages if `True` (default)
        :raises ValueError: if there are too many messages or the positions indices are
            invalid.
        """
        # validate inputs
        if len(msgs) > self.message_len:
            raise ValueError(
                f"Too many message: {len(msgs)} given, max. {self.message_len} allowed."
            )

        timestamp = datetime.now().time().strftime("%H:%M:%S")
        # append messages in the same order as given, not reversed
        self.message_board.extendleft(f"{timestamp}: {msg}" for msg in reversed(msgs))

        if display_board:
            self.display_message_board()

    def display_message_board(self) -> None:
        """
        Display 15 newest messages
        """
        return self._display_messages(self.message_board)

    def _display_messages(self, messages: Sequence[str]) -> None:
        """
        Display given messages on message board

        :param messages: sequence of messages to display
        """
        # Note: cannot zip(constants.MessageBoard, messages) as enum instances are
        #       sorted by name, hence after after `line_1` comes `line_10`, not `line_2`
        for n, msg in enumerate(messages):
            line = constants.MessageBoard.line(n + 1)
            self.write(line, msg)



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Test module for the Supercube Variant A class.
"""

import pytest

from hvl_ccb.dev.supercube2015 import Supercube2015WithFU, constants
from tests.opctools import DemoServer


@pytest.fixture(scope='module')
def dev_config():
    return {
        'namespace_index': 7,
    }


@pytest.fixture(scope='module')
def com_config():
    return {
        'host': 'localhost',
        'port': 14224,
        'endpoint_name': ''
    }


@pytest.fixture(scope='module')
def demo_server(dev_config):
    opcua_server = DemoServer(dev_config['namespace_index'], 'A', 14224)
    opcua_server.start()

    # add socket nodes
    for socket in constants.GeneralSockets:
        opcua_server.add_var(socket, False, True)

    # add support input and output nodes
    for support in constants.GeneralSupport:
        opcua_server.add_var(support, False, True)

    for safety in constants.Safety:
        opcua_server.add_var(safety, False, True)

    for stick in constants.EarthingStick:
        opcua_server.add_var(stick, False, True)

    opcua_server.add_var(constants.BreakdownDetection.activated, False, False)
    opcua_server.add_var(constants.BreakdownDetection.triggered, False, False)

    yield opcua_server

    opcua_server.stop()


@pytest.fixture(scope='module')
def cube(com_config, dev_config, demo_server):
    cube = Supercube2015WithFU(com_config, dev_config)
    cube.start()
    yield cube
    cube.stop()


def test_instantiation(com_config):
    cube = Supercube2015WithFU(com_config)
    assert cube is not None


def test_set_voltage(cube: Supercube2015WithFU, demo_server: DemoServer):
    demo_server.add_var(constants.Power.voltage_target, 0.0, True)
    cube.set_target_voltage(20e3)
    assert demo_server.get_var(constants.Power.voltage_target) == 20.0


def test_get_voltage(cube: Supercube2015WithFU, demo_server: DemoServer):
    cube.set_target_voltage(1e3)
    assert cube.get_target_voltage() == 1e3


def test_get_max_voltage(cube: Supercube2015WithFU, demo_server: DemoServer):
    demo_server.add_var(constants.Power.voltage_max, 80, False)
    assert cube.get_max_voltage() == 80.0e3


def test_set_slope(cube: Supercube2015WithFU, demo_server: DemoServer):
    demo_server.add_var(constants.Power.voltage_slope, 1.5, True)
    cube.set_slope(5.5e3)
    assert demo_server.get_var(constants.Power.voltage_slope) == 5.5


def test_get_primary_voltage(cube: Supercube2015WithFU, demo_server: DemoServer):
    demo_server.add_var(constants.Power.voltage_primary, 23.2, False)
    assert cube.get_primary_voltage() == 23.2


def test_get_primary_current(cube: Supercube2015WithFU, demo_server: DemoServer):
    demo_server.add_var(constants.Power.current_primary, 1.2, False)
    assert cube.get_primary_current() == 1.2


def test_get_frequency(cube: Supercube2015WithFU, demo_server: DemoServer):
    demo_server.add_var(constants.Power.frequency, 50, False)
    assert cube.get_frequency() == 50.0


def test_get_power_setup(cube: Supercube2015WithFU, demo_server: DemoServer):
    demo_server.add_var(constants.Power.setup, 1, False)
    assert cube.get_power_setup() == constants.PowerSetup(1)


def test_get_fso_active(cube: Supercube2015WithFU, demo_server: DemoServer):
    demo_server.set_var(constants.BreakdownDetection.activated, True)
    assert cube.get_fso_active() is True


def test_fso_reset(cube: Supercube2015WithFU, demo_server: DemoServer):
    cube.fso_reset()
    assert demo_server.get_var(constants.BreakdownDetection.reset) is False



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Constants, variable names for the Supercube OPC-connected devices.
"""

import logging
from typing import cast, Sequence, Sized, Tuple

from aenum import IntEnum, EnumMeta

from hvl_ccb.utils.enum import ValueEnum, unique

logger = logging.getLogger(__name__)


@unique
class SupercubeOpcEndpoint(ValueEnum):
    """
    OPC Server Endpoint strings for the supercube variants.
    """

    A = "Supercube Typ A"
    B = "Supercube Typ B"


# NOTE: super metaclass has to match metaclass of `super(GeneralSupport)`!
class GeneralSupportMeta(EnumMeta):
    def __new__(metacls, clsname, bases, clsdict, **kwargs):
        cls = EnumMeta.__new__(metacls, clsname, bases, clsdict, **kwargs)
        # BEWARE: keep manually in sync with suffixes of enum instances; opt use dir()
        cls._port_range = range(1, 7)
        cls._contact_range = range(1, 3)
        return cls


@unique
class GeneralSupport(ValueEnum, metaclass=GeneralSupportMeta):
    """
    NodeID strings for the support inputs and outputs.
    """

    in_1_1 = '"Ix_Allg_Support1_1"'
    in_1_2 = '"Ix_Allg_Support1_2"'
    in_2_1 = '"Ix_Allg_Support2_1"'
    in_2_2 = '"Ix_Allg_Support2_2"'
    in_3_1 = '"Ix_Allg_Support3_1"'
    in_3_2 = '"Ix_Allg_Support3_2"'
    in_4_1 = '"Ix_Allg_Support4_1"'
    in_4_2 = '"Ix_Allg_Support4_2"'
    in_5_1 = '"Ix_Allg_Support5_1"'
    in_5_2 = '"Ix_Allg_Support5_2"'
    in_6_1 = '"Ix_Allg_Support6_1"'
    in_6_2 = '"Ix_Allg_Support6_2"'
    out_1_1 = '"Qx_Allg_Support1_1"'
    out_1_2 = '"Qx_Allg_Support1_2"'
    out_2_1 = '"Qx_Allg_Support2_1"'
    out_2_2 = '"Qx_Allg_Support2_2"'
    out_3_1 = '"Qx_Allg_Support3_1"'
    out_3_2 = '"Qx_Allg_Support3_2"'
    out_4_1 = '"Qx_Allg_Support4_1"'
    out_4_2 = '"Qx_Allg_Support4_2"'
    out_5_1 = '"Qx_Allg_Support5_1"'
    out_5_2 = '"Qx_Allg_Support5_2"'
    out_6_1 = '"Qx_Allg_Support6_1"'
    out_6_2 = '"Qx_Allg_Support6_2"'

    @classmethod
    def port_range(cls) -> Sequence[int]:
        """
        Integer range of all ports.

        :return: sequence of port numbers
        """
        return cls._port_range

    @classmethod
    def contact_range(cls) -> Sequence[int]:
        """
        Integer range of all contacts.

        :return: sequence of contact numbers
        """
        return cls._contact_range

    @classmethod
    def _validate_port_contact_numbers(cls, port: int, contact: int):
        """
        Validate earthing stick number.

        :param port: the port number
        :param contact: the contact number
        :raises ValueError: when port or contact number is not valid
        """
        if port not in cls.port_range():
            raise ValueError(f"Port number must be one of {list(cls.port_range())}")
        if contact not in cls.contact_range():
            raise ValueError(
                f"Contact number must be one of {list(cls.contact_range())}"
            )

    @classmethod
    def output(cls, port: int, contact: int):
        """
        Get the NodeID string for a support output.

        :param port: the desired port (1..6)
        :param contact: the desired contact at the port (1..2)
        :return: the node id string
        :raises ValueError: when port or contact number is not valid
        """
        cls._validate_port_contact_numbers(port, contact)
        return getattr(cls, f"out_{port}_{contact}")

    @classmethod
    def input(cls, port: int, contact: int):
        """
        Get the NodeID string for a support input.

        :param port: the desired port (1..6)
        :param contact: the desired contact at the port (1..2)
        :return: the node id string
        :raises ValueError: when port or contact number is not valid
        """
        cls._validate_port_contact_numbers(port, contact)
        return getattr(cls, f"in_{port}_{contact}")


@unique
class BreakdownDetection(ValueEnum):
    """
    Node ID strings for the breakdown detection.

    TODO: these variable NodeIDs are not tested and/or correct yet.
    """

    #: Boolean read-only variable indicating whether breakdown detection and fast
    #: switchoff is enabled in the system or not.
    activated = '"Ix_Allg_Breakdown_activated"'

    #: Boolean read-only variable telling whether the fast switch-off has triggered.
    #: This can also be seen using the safety circuit state, therefore no method is
    #: implemented to read this out directly.
    triggered = '"Ix_Allg_Breakdown_triggered"'

    #: Boolean writable variable to reset the fast switch-off. Toggle to re-enable.
    reset = '"Qx_Allg_Breakdown_reset"'


@unique
class GeneralSockets(ValueEnum):
    """
    NodeID strings for the power sockets (3x T13 and 1xCEE16).
    """

    #: SEV T13 socket No. 1 (writable boolean).
    t13_1 = '"Qx_Allg_Socket_T13_1"'

    #: SEV T13 socket No. 2 (writable boolean).
    t13_2 = '"Qx_Allg_Socket_T13_2"'

    #: SEV T13 socket No. 3 (writable boolean).
    t13_3 = '"Qx_Allg_Socket_T13_3"'

    #: CEE16 socket (writeable boolean).
    cee16 = '"Qx_Allg_Socket_CEE16"'


T13_SOCKET_PORTS = (1, 2, 3)
"""
Port numbers of SEV T13 power socket
"""


@unique
class Safety(ValueEnum):
    """
    NodeID strings for the basic safety circuit status and green/red switches "ready"
    and "operate".
    """

    #: Status is a read-only integer containing the state number of the
    #: supercube-internal state machine. The values correspond to numbers in
    #: :class:`SafetyStatus`.
    status = '"DB_Safety_Circuit"."si_safe_status"'

    #: Writable boolean for switching to Red Ready (locked, HV off) state.
    switch_to_ready = '"DB_Safety_Circuit"."sx_safe_switch_to_ready"'

    #: Writable boolean for switching to Red Operate (locket, HV on) state.
    switch_to_operate = '"DB_Safety_Circuit"."sx_safe_switch_to_operate"'


class SafetyStatus(IntEnum):
    """
    Safety status values that are possible states returned from
    :meth:`hvl_ccb.dev.supercube.base.Supercube.get_status`. These
    values correspond to the states of the Supercube's safety circuit statemachine.
    """

    #: System is initializing or booting.
    Initializing = 0

    #: System is safe, lamps are green and some safety elements are not in place such
    #: that it cannot be switched to red currently.
    GreenNotReady = 1

    #: System is safe and all safety elements are in place to be able to switch to
    #: *ready*.
    GreenReady = 2

    #: System is locked in red state and *ready* to go to *operate* mode.
    RedReady = 3

    #: System is locked in red state and in *operate* mode, i.e. high voltage on.
    RedOperate = 4

    #: Fast turn off triggered and switched off the system. Reset FSO to go back to a
    #: normal state.
    QuickStop = 5

    #: System is in error mode.
    Error = 6


@unique
class Power(ValueEnum):
    """
    Variable NodeID strings concerning power data.

    TODO: these variable NodeIDs are not tested and/or correct yet, they don't exist
        yet on Supercube side.
    """

    #: Primary voltage in volts, measured by the frequency converter at its output.
    #: (read-only)
    voltage_primary = "Qr_Power_FU_actual_Voltage"

    #: Primary current in ampere, measured by the frequency converter. (read-only)
    current_primary = "Qr_Power_FU_actual_Current"

    #: Power setup that is configured using the Supercube HMI. The value corresponds to
    #: the ones in :class:`PowerSetup`. (read-only)
    setup = "Qi_Power_Setup"

    #: Voltage slope in V/s.
    voltage_slope = "Ir_Power_dUdt"

    #: Target voltage setpoint in V.
    voltage_target = "Ir_Power_Target_Voltage"

    #: Maximum voltage allowed by the current experimental setup. (read-only)
    voltage_max = "Iw_Power_max_Voltage"

    #: Frequency converter output frequency. (read-only)
    frequency = "Ir_Power_FU_Frequency"


class PowerSetup(IntEnum):
    """
    Possible power setups corresponding to the value of variable :attr:`Power.setup`.
    """

    #: No safety switches, use only safety components (doors, fence, earthing...)
    #: without any power.
    NoPower = 0

    #: External power supply fed through blue CEE32 input using isolation transformer
    #: and safety switches of the Supercube, or using an external safety switch
    #: attached to the Supercube Type B.
    External = 1

    #: AC voltage with MWB transformer set to 50kV maximum voltage.
    AC_SingleStage_50kV = 2

    #: AC voltage with MWB transformer set to 100kV maximum voltage.
    AC_SingleStage_100kV = 3

    #: AC voltage with two MWB transformers, one at 100kV and the other at 50kV,
    #: resulting in a total maximum voltage of 150kV.
    AC_DoubleStage_150kV = 4

    #: AC voltage with two MWB transformers both at 100kV, resulting in a total
    #: maximum voltage of 200kV
    AC_DoubleStage_200kV = 5

    #: Internal usage of the frequency converter, controlling to the primary voltage
    #: output of the supercube itself (no measurement transformer used)
    Internal = 6

    #: DC voltage with one AC transformer set to 100kV AC, resulting in 140kV DC
    DC_SingleStage_140kV = 7

    #: DC voltage with two AC transformers set to 100kV AC each, resulting in 280kV
    #: DC in total (or a single stage transformer with Greinacher voltage doubling
    #: rectifier)
    DC_DoubleStage_280kV = 8


class _PrefixedNumbersEnumBase(ValueEnum):
    """
    Base class for enums with "{prefix}{n}" instance names, where n=1..N.
    """

    @classmethod
    def range(cls) -> Sequence[int]:
        """
        Integer range of all channels.

        :return: sequence of channel numbers
        """
        return range(1, len(cast(Sized, cls)) + 1)

    @classmethod
    def _validate_number(cls, number: int):
        """
        Validate enum instance number.

        :param number: the enum instance number (1..N)
        :raises ValueError: when enum instance number is not in 1..N range
        """
        if number not in cls.range():
            raise ValueError(
                f"{cls._prefix()} number must be one of {list(cls.range())}"
            )

    @classmethod
    def _prefix(cls) -> str:
        """
        Enum instances name prefix: "{prefix}{n}"

        :return: enum instances prefix string
        """
        raise NotImplementedError("Implement in subclass")

    @property
    def number(self) -> int:
        """
        Get corresponding enum instance number.

        :return: enum instance number (1..N)
        """
        # Py >=3.9: self.name.removeprefix()
        return int(self.name[len(self._prefix()):])

    # no type return as it would be a arguably too complex/obscure;
    # cf. https://github.com/python/typing/issues/58#issuecomment-326240794
    @classmethod
    def get(cls, number: int):
        """
        Get the enum instance for a given number.

        :param number: the instance number (1..N)
        :return: the enum instance for the given number.
        :raises ValueError: when instance number is not in the 1..N range
        """
        cls._validate_number(number)
        return getattr(cls, f"{cls._prefix()}{number}")


class _InputEnumBase(_PrefixedNumbersEnumBase):
    """
    Base class for enums with "input_{n}" instance names, where n=1..N.
    """

    @classmethod
    def _prefix(cls) -> str:
        return "input_"

    @classmethod
    def input(cls, channel: int):
        """
        Get the enum instance for a given channel number.

        :param channel: the channel number (1..N)
        :return: the enum instance for the given channel.
        :raises ValueError: when channel number is not in the 1..N range
        """
        return cls.get(channel)


MeasurementsScaledInput = unique(
    _InputEnumBase(
        "MeasurementsScaledInput",
        {
            f"{_InputEnumBase._prefix()}{n}":
                f'"DB_Measurements"."si_scaled_Voltage_Input_{n}"'
            for n in range(1, 5)
        },
    )
)
"""
Variable NodeID strings for the four analog BNC inputs for measuring voltage.
The voltage returned in these variables is already scaled with the set ratio,
which can be read using the variables in :class:`MeasurementsDividerRatio`.
"""


MeasurementsDividerRatio = unique(
    _InputEnumBase(
        "MeasurementsDividerRatio",
        {
            f"{_InputEnumBase._prefix()}{n}":
                f'"DB_Measurements"."si_Divider_Ratio_{n}"'
            for n in range(1, 5)
        },
    )
)
"""
Variable NodeID strings for the measurement input scaling ratios. These ratios
are defined in the Supercube HMI setup and are provided in the python module here
to be able to read them out, allowing further calculations.
"""


class EarthingStickStatus(IntEnum):
    """
    Status of an earthing stick. These are the possible values in the status integer
    e.g. in :attr:`EarthingStick.status_1`.
    """

    #: Earthing stick is deselected and not enabled in safety circuit. To get out of
    #: this state, the earthing has to be enabled in the Supercube HMI setup.
    inactive = 0

    #: Earthing is closed (safe).
    closed = 1

    #: Earthing is open (not safe).
    open = 2

    #: Earthing is in error, e.g. when the stick did not close correctly or could not
    #: open.
    error = 3


class EarthingStickOperatingStatus(IntEnum):
    """
    Operating Status for an earthing stick. Stick can be used in auto or manual mode.
    """

    auto = 0
    manual = 1


class EarthingStickOperation(IntEnum):
    """
    Operation of the earthing stick in manual operating mode. Can be closed of opened.
    """

    open = 0
    close = 1


# NOTE: super metaclass has to match metaclass of `super(EarthingStick)`!
class EarthingStickMeta(EnumMeta):
    def __new__(metacls, clsname, bases, clsdict, **kwargs):
        cls = EnumMeta.__new__(metacls, clsname, bases, clsdict, **kwargs)
        # BEWARE: keep manually in sync with suffixes of enum instances; opt use dir()
        cls._range = range(1, 7)
        cls._statuses = tuple(getattr(cls, f"status_{number}") for number in cls._range)
        cls._manuals = tuple(getattr(cls, f"manual_{number}") for number in cls._range)
        cls._operating_statuses = tuple(
            getattr(cls, f"operating_status_{number}") for number in cls._range
        )
        return cls


@unique
class EarthingStick(ValueEnum, metaclass=EarthingStickMeta):
    """
    Variable NodeID strings for all earthing stick statuses (read-only integer) and
    writable booleans for setting the earthing in manual mode.
    """

    status_1 = '"DB_Safety_Circuit"."Earthstick_1"."si_HMI_Status"'
    status_2 = '"DB_Safety_Circuit"."Earthstick_2"."si_HMI_Status"'
    status_3 = '"DB_Safety_Circuit"."Earthstick_3"."si_HMI_Status"'
    status_4 = '"DB_Safety_Circuit"."Earthstick_4"."si_HMI_Status"'
    status_5 = '"DB_Safety_Circuit"."Earthstick_5"."si_HMI_Status"'
    status_6 = '"DB_Safety_Circuit"."Earthstick_6"."si_HMI_Status"'

    manual_1 = '"DB_Safety_Circuit"."Earthstick_1"."sx_earthing_manually"'
    manual_2 = '"DB_Safety_Circuit"."Earthstick_2"."sx_earthing_manually"'
    manual_3 = '"DB_Safety_Circuit"."Earthstick_3"."sx_earthing_manually"'
    manual_4 = '"DB_Safety_Circuit"."Earthstick_4"."sx_earthing_manually"'
    manual_5 = '"DB_Safety_Circuit"."Earthstick_5"."sx_earthing_manually"'
    manual_6 = '"DB_Safety_Circuit"."Earthstick_6"."sx_earthing_manually"'

    operating_status_1 = '"DB_Safety_Circuit"."Earthstick_1"."sx_manual_control_active"'
    operating_status_2 = '"DB_Safety_Circuit"."Earthstick_2"."sx_manual_control_active"'
    operating_status_3 = '"DB_Safety_Circuit"."Earthstick_3"."sx_manual_control_active"'
    operating_status_4 = '"DB_Safety_Circuit"."Earthstick_4"."sx_manual_control_active"'
    operating_status_5 = '"DB_Safety_Circuit"."Earthstick_5"."sx_manual_control_active"'
    operating_status_6 = '"DB_Safety_Circuit"."Earthstick_6"."sx_manual_control_active"'

    @property
    def number(self) -> int:
        """
        Get corresponding earthing stick number.

        :return: earthing stick number (1..6)
        """
        return int(self.name.rpartition("_")[-1])

    @classmethod
    def range(cls) -> Sequence[int]:
        """
        Integer range of all earthing sticks.

        :return: sequence of earthing sticks numbers
        """
        return cls._range

    @classmethod
    def _validate_earthing_stick_number(cls, number: int):
        """
        Validate earthing stick number.

        :param number: the earthing stick number
        :raises ValueError: when earthing stick number is not valid
        """
        if number not in cls.range():
            raise ValueError(
                f"Earthing stick number must be one of {list(cls.range())}"
            )

    @classmethod
    def statuses(cls) -> Tuple["EarthingStick", ...]:
        """
        Get all earthing stick status instances.

        :return: tuple of status instances
        """
        return cls._statuses

    @classmethod
    def status(cls, number: int) -> "EarthingStick":
        """
        Get the status enum instance for an earthing stick number.

        :param number: the earthing stick (1..6)
        :return: the status instance
        :raises ValueError: when earthing stick number is not valid
        """
        cls._validate_earthing_stick_number(number)
        return cls.statuses()[number - 1]

    @classmethod
    def manuals(cls) -> Tuple["EarthingStick", ...]:
        """
        Get all earthing stick manual instances.

        :return: tuple of manual instances
        """
        return cls._manuals

    @classmethod
    def manual(cls, number: int):
        """
        Get the manual enum instance for an earthing stick number.

        :param number: the earthing stick (1..6)
        :return: the manual instance
        :raises ValueError: when earthing stick number is not valid
        """
        cls._validate_earthing_stick_number(number)
        return cls.manuals()[number - 1]

    @classmethod
    def operating_statuses(cls) -> Tuple["EarthingStick", ...]:
        """
        Get all earthing stick operating status instances.

        :return: tuple of operating status instances
        """
        return cls._operating_statuses

    @classmethod
    def operating_status(cls, number: int):
        """
        Get the operating status enum instance for an earthing stick number.

        :param number: the earthing stick (1..6)
        :return: the operating status instance
        :raises ValueError: when earthing stick number is not valid
        """
        cls._validate_earthing_stick_number(number)
        return cls.operating_statuses()[number - 1]


@unique
class Errors(ValueEnum):
    """
    Variable NodeID strings for information regarding error, warning and message
    handling.
    """

    #: Boolean read-only variable telling if a message is active.
    message = '"DB_Message_Buffer"."Info_active"'

    #: Boolean read-only variable telling if a warning is active.
    warning = '"DB_Message_Buffer"."Warning_active"'

    #: Boolean read-only variable telling if a stop is active.
    stop = '"DB_Message_Buffer"."Stop_active"'

    #: Writable boolean for the error quit button.
    quit = '"DB_Message_Buffer"."Reset_button"'


class _AlarmEnumBase(_PrefixedNumbersEnumBase):
    """
    Base class for enums with "Alarm{n}" instance names, where n=1..N.
    """

    @classmethod
    def _prefix(cls) -> str:
        return "Alarm"

    @classmethod
    def Alarm(cls, number: int):
        """
        Get the enum instance for a given alarm number.

        :param number: the alarm number (1..N)
        :return: the enum instance for the alarm number.
        :raises ValueError: when alarm number is not in the 1..N range
        """
        return cls.get(number)


Alarms = unique(
    _AlarmEnumBase(
        "Alarms",
        {f"Alarm{n}": f'"DB_Alarm_HMI"."Alarm{n}"' for n in range(1, 152)},
    )
)
"""
Alarms enumeration containing all variable NodeID strings for the alarm array.
"""


class DoorStatus(IntEnum):
    """
    Possible status values for doors.
    """

    #: not enabled in Supercube HMI setup, this door is not supervised.
    inactive = 0

    #: Door is open.
    open = 1

    #: Door is closed, but not locked.
    closed = 2

    #: Door is closed and locked (safe state).
    locked = 3

    #: Door has an error or was opened in locked state (either with emergency stop or
    #: from the inside).
    error = 4


class EarthingRodStatus(IntEnum):
    """
    Possible status values for earthing rods.
    """

    #: earthing rod is somewhere in the experiment
    #: and blocks the start of the experiment
    experiment_blocked = 0

    #: earthing rod is hanging next to the door, experiment is ready to operate
    experiment_ready = 1


class _DoorEnumBase(_PrefixedNumbersEnumBase):
    """
    Base class for enums with "status_{n}" instance names, where n=1..N is the door
    number.
    """

    @classmethod
    def _prefix(cls) -> str:
        return "status_"

    @classmethod
    def status(cls, door: int):
        """
        Get the enum instance for a given door number.

        :param number: the door number (1..N)
        :return: the enum instance for the given door number.
        :raises ValueError: when door number is not in the 1..N range
        """
        return cls.get(door)


Door = unique(
    _DoorEnumBase(
        "Door",
        {
            f"{_DoorEnumBase._prefix()}{n}":
                f'"DB_Safety_Circuit"."Door_{n}"."si_HMI_status"'
            for n in range(1, 4)
        },
    )
)
"""
Variable NodeID strings for doors.
"""

EarthingRod = unique(
    _DoorEnumBase(
        "EarthingRod",
        {
            f"{_DoorEnumBase._prefix()}{n}":
                f'"DB_Safety_Circuit"."Door_{n}"."Ix_earthingrod"'
            for n in range(1, 4)
        },
    )
)
"""
Variable NodeID strings for earthing rods.
"""


class AlarmText(ValueEnum):
    """
    This enumeration contains textual representations for all error classes (stop,
    warning and message) of the Supercube system. Use the :meth:`AlarmText.get`
    method to retrieve the enum of an alarm number.
    """

    # Safety elements
    Alarm1 = "STOP Emergency Stop 1"
    Alarm2 = "STOP Emergency Stop 2"
    Alarm3 = "STOP Emergency Stop 3"
    Alarm4 = "STOP Safety Switch 1 error"
    Alarm5 = "STOP Safety Switch 2 error"
    Alarm6 = "STOP Door 1 lock supervision"
    Alarm7 = "STOP Door 2 lock supervision"
    Alarm8 = "STOP Door 3 lock supervision"
    Alarm9 = "STOP Earthing stick 1 error while opening"
    Alarm10 = "STOP Earthing stick 2 error while opening"
    Alarm11 = "STOP Earthing stick 3 error while opening"
    Alarm12 = "STOP Earthing stick 4 error while opening"
    Alarm13 = "STOP Earthing stick 5 error while opening"
    Alarm14 = "STOP Earthing stick 6 error while opening"
    Alarm15 = "STOP Earthing stick 1 error while closing"
    Alarm16 = "STOP Earthing stick 2 error while closing"
    Alarm17 = "STOP Earthing stick 3 error while closing"
    Alarm18 = "STOP Earthing stick 4 error while closing"
    Alarm19 = "STOP Earthing stick 5 error while closing"
    Alarm20 = "STOP Earthing stick 6 error while closing"
    Alarm21 = "STOP Safety fence 1"
    Alarm22 = "STOP Safety fence 2"
    Alarm23 = "STOP OPC connection error"
    Alarm24 = "STOP Grid power failure"
    Alarm25 = "STOP UPS failure"
    Alarm26 = "STOP 24V PSU failure"

    # Doors
    Alarm41 = "WARNING Door 1: Use earthing rod!"
    Alarm42 = "MESSAGE Door 1: Earthing rod is still in setup."
    Alarm43 = "WARNING Door 2: Use earthing rod!"
    Alarm44 = "MESSAGE Door 2: Earthing rod is still in setup."
    Alarm45 = "WARNING Door 3: Use earthing rod!"
    Alarm46 = "MESSAGE Door 3: Earthing rod is still in setup."

    # General
    Alarm47 = "MESSAGE UPS charge < 85%"
    Alarm48 = "MESSAGE UPS running on battery"

    # generic not defined alarm text
    not_defined = "NO ALARM TEXT DEFINED"

    @classmethod
    def get(cls, alarm: int):
        """
        Get the attribute of this enum for an alarm number.

        :param alarm: the alarm number
        :return: the enum for the desired alarm number
        """

        try:
            return getattr(cls, f"Alarm{alarm}")
        except AttributeError:
            return cls.not_defined


class OpcControl(ValueEnum):
    """
    Variable NodeID strings for supervision of the OPC connection from the
    controlling workstation to the Supercube.
    """

    #: writable boolean to enable OPC remote control and display a message window on
    #: the Supercube HMI.
    active = '"DB_OPC_Connection"."sx_OPC_active"'
    live = '"DB_OPC_Connection"."sx_OPC_lifebit"'


class _LineEnumBase(_PrefixedNumbersEnumBase):
    """
    Base class for enums with "input_{n}" instance names, where n=1..N.
    """

    @classmethod
    def _prefix(cls) -> str:
        return "line_"

    @classmethod
    def line(cls, number: int):
        """
        Get the enum instance for a given line number.

        :param number: the line number (1..M)
        :return: the enum instance for the given line number.
        :raises ValueError: when line number is not in the 1..N range
        """
        return cls.get(number)


MessageBoard = unique(
    _LineEnumBase(
        "MessageBoard",
        {
            f"{_LineEnumBase._prefix()}{n}": f'"DB_OPC_Connection"."Is_status_Line_{n}"'
            for n in range(1, 16)
        },
    )
)
"""
Variable NodeID strings for message board lines.
"""



#!/usr/bin/env python
#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'pyserial>=3.4',
    'pymodbus>=2.3.0',
    'IPy>=0.83',
    'bitstring>=3.1.5',
    'pyvisa>=1.11.3',
    'pyvisa-py>=0.5.2',
    'typeguard>=2.3.0',
    'aenum>=2.1.2',
    'opcua>=0.98.6',
    'cryptography>=2.6.1',  # optional dependency of the opcua package
    'openpyxl>=2.6.2',
]

tiepie = ['python-libtiepie>=0.7', 'numpy>=1.19.0']
labjack = ['labjack-ljm>=1.20.0']
all_libs = tiepie + labjack

extra_requirements = {
    'all': all_libs,
    'tiepie': tiepie,
    'labjack': labjack,
}


dependency_links = []

setup(
    author=(
        'Mikołaj Rybiński, David Graber, Henrik Menne, Alise Chachereau, '
        'Henning Janssen',
    ),
    author_email=(
        'mikolaj.rybinski@id.ethz.ch, dev@davidgraber.ch, henrik.menne@eeh.ee.ethz.ch, '
        'chachereau@eeh.ee.ethz.ch, janssen@eeh.ee.ethz.ch'
    ),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description=(
        'Python common code base to control devices high voltage research devices, in '
        'particular, as used in Christian Franck\'s High Voltage Lab (HVL), D-ITET, ETH'
    ),
    entry_points={},
    install_requires=requirements,
    extras_require=extra_requirements,
    dependency_links=dependency_links,
    license='GNU General Public License v3',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    keywords='hvl_ccb',
    name='hvl_ccb',
    package_data={
        'hvl_ccb': ['py.typed'],
    },
    packages=find_packages(),
    test_suite='tests',
    url='https://gitlab.com/ethz_hvl/hvl_ccb/',
    version='0.7.0',
    zip_safe=False,
)



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for VisaDevice
"""
from time import sleep

import pytest

from hvl_ccb.comm import (
    CommunicationProtocol,
    VisaCommunicationConfig
)
from hvl_ccb.dev import VisaDevice
from tests.masked_comm import MaskedVisaCommunication


@pytest.fixture(scope='module')
def com_config():
    return VisaCommunicationConfig(
        interface_type=VisaCommunicationConfig.InterfaceType.TCPIP_SOCKET,
        board=0,
        host='192.168.1.31',
        port=5025
    )


@pytest.fixture(scope='module')
def dev_config():
    return {
        'spoll_interval': 0.01,
        'spoll_start_delay': 0.01,
    }


class WrongCommunicationProtocol(CommunicationProtocol):
    @staticmethod
    def config_cls():
        return VisaCommunicationConfig

    def open(self):
        pass

    def close(self):
        pass


def test_instantiation(com_config, dev_config):
    dev = VisaDevice(com_config, dev_config)

    assert com_config.address == "TCPIP0::192.168.1.31::5025::SOCKET"
    assert dev is not None

    dev_config_dict = dev_config

    assert dev.config.spoll_interval == dev_config_dict['spoll_interval']

    invalid_dev_config = dict(dev_config_dict)
    invalid_dev_config['spoll_interval'] = 0
    with pytest.raises(ValueError):
        VisaDevice(com_config, invalid_dev_config)

    assert dev.config.spoll_start_delay == dev_config_dict['spoll_start_delay']

    invalid_dev_config = dict(dev_config_dict)
    invalid_dev_config['spoll_start_delay'] = -1
    with pytest.raises(ValueError):
        VisaDevice(com_config, invalid_dev_config)


def test_basics(com_config, dev_config):
    com = MaskedVisaCommunication(com_config)
    dev = VisaDevice(com, dev_config)

    com.put_name('*IDN?', 'test_identification')
    assert dev.get_identification() == 'test_identification'

    com.put_name('SYSTem:ERRor:ALL?', '0,"No error"')
    assert dev.get_error_queue() == '0,"No error"'

    dev.reset()
    assert com.get_written() == '*RST'
    assert com.get_written() == '*CLS'

    dev.start()

    for i in range(8):
        com.stb = 2**i - 1
        sleep(0.1)

    dev.stop()

    assert com == dev.com



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Test module for the Supercube Variant B class.
"""

import pytest

from hvl_ccb.dev.supercube import SupercubeB


@pytest.fixture(scope='module')
def com_config():
    return {
        'host': '127.0.0.1',
    }


def test_instantiation(com_config):
    cube = SupercubeB(com_config)
    assert cube is not None



#  Copyright (c) 2019-2021 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tools for OPC testing
"""

from opcua import Server
from opcua.ua import NodeId, Variant


class DemoServer(Server):

    def __init__(self, namespace, cube_type: str, port: int = 4840, shelffile=None,
                 iserver=None):
        super().__init__(shelffile, iserver)

        self.set_endpoint(f'opc.tcp://0.0.0.0:{port}/freeopcua/server/')

        self._ns = namespace
        self._cube_type = cube_type

        self._root = self.get_objects_node().add_object(
            self._ns, f'Supercube {self._cube_type}'
        )

    def add_var(self, id, val, writable: bool):
        var = self._root.add_variable(
            NodeId(identifier=str(id),
                   namespaceidx=self._ns),
            str(id),
            Variant(val)
        )
        var.set_writable(writable)
        return var

    def set_var(self, id, val):
        self.get_node(
            NodeId(identifier=str(id),
                   namespaceidx=self._ns),
        ).set_value(Variant(val))

    def get_var(self, id):
        return self.get_node(
            NodeId(identifier=str(id),
                   namespaceidx=self._ns),
        ).get_value()
